import copy

from .goal_metadata import GoalMetadata
from .return_status import ReturnStatus
from .stringifier import Stringifier


class StepMetadata:
    """
    Defines output metadata from an action step in the MCS 3D environment.

    Attributes
    ----------
    action_list : list of (string, dict) tuples
        The list of all actions that are available for the next step.
        Each action is returned as a tuple containing the action string and
        the action's restricted parameters, if any.
        For example: ("Pass", {}) forces a Pass action; ("PickupObject", {})
        forces a PickupObject action with any parameters; and
        ("PickupObject", {"objectId": "a"}) forces a PickupObject action with
        the specific parameters objectId=a.
        EndHabituation is a special case of the action_list where its
        parameters will always be empty. When taking the EndHabituation
        action, the MCS system may apply hidden displacement parameters to the
        robot.
        An action_list of None or an empty list means that all actions will
        be available for the next step.

        To "step" using the first action from the action_list:

        .. highlight:: python
        .. code-block:: python

            step_metadata = controller.start_scene(scene_data)
            action, params = step_metadata.action_list[0]
            step_metadata = controller.step(action, **params)

        Derived from :mod:`GoalMetadata.action_list[step_number]
        <machine_common_sense.GoalMetadata>`.
        May be a subset of all possible actions. See
        :mod:`Action <machine_common_sense.Action>`.
    camera_aspect_ratio : (float, float)
        The player camera's aspect ratio. This will remain constant for the
        whole scene.
    camera_clipping_planes : (float, float)
        The player camera's near and far clipping planes, in meters. This will
        remain constant for the whole scene. Default (0.01, 150)
    camera_field_of_view : float
        The player camera's field of view. This will remain constant for
        the whole scene.
    camera_height : float
        The player camera's height, in meters.
    depth_map_list : list of 2D numpy arrays
        The list of 2-dimensional numpy arrays of depth float data from the
        scene after the last action and physics simulation were run. This is
        usually a list with 1 array, except for the output from start_scene
        for a scene with a scripted Preview Phase (Preview Phase case details
        TBD).
        Each 32-bit depth float in the 2-dimensional numpy array is a value
        between the camera's near clipping plane (default 0.01) and the
        camera's far clipping plane (default 150) corresponding to the depth,
        in meters, at that pixel in the image.
        Note that this list will be empty if the metadata level is 'none'.
    goal : GoalMetadata or None
        The goal for the whole scene. Will be None in "Exploration" scenes.
    haptic_feedback : dict
        Haptic feedback sources for the agent. Values are true or false
        depending on if the agent is touching the haptic feedback source.
        The only current supported contact is "on_lava"
    habituation_trial : int or None
        The current habituation trial (as a positive integer), or None if the
        scene is not currently in a habituation trial (meaning this scene is
        in a test trial).
    head_tilt : float
        How far your head is tilted up/down in degrees (between 90 and -90).
        Changed by setting the "horizon" parameter in a "RotateLook" action.
    holes : list of tuples
        Coordinates of holes as (X, Z) float tuples. Will be set to 'None' if
        using a metadata level below the 'oracle' level.
    image_list : list of Pillow.Image objects
        The list of images from the scene after the last action and physics
        simulation were run. This is usually a list with 1 image, except for
        the output from start_scene for a scene with a scripted Preview Phase.
        (Preview Phase case details TBD).
    lava : list of tuples
        Coordinates of pools of lava as (X1, Z1, X2, Z2) float tuples, where
        X1/Z1 is the top-left corner and X2/Z2 is the bottom-right conrer. Will
        be set to 'None' if using a metadata level below the 'oracle' level.
    object_list : list of ObjectMetadata objects
        The list of metadata for all the visible interactive objects in the
        scene. This list will be empty if using a metadata level below
        the 'oracle' level. For metadata on structural objects like walls,
        please see structural_object_list
    object_mask_list : list of Pillow.Image objects
        The list of object mask (instance segmentation) images from the scene
        after the last action and physics simulation were run. This is usually
        a list with 1 image, except for the output from start_scene for a
        scene with a scripted Preview Phase (Preview Phase case details TBD).
        The color of each object in the mask corresponds to the "color"
        property in its ObjectMetadata object.
        Note that this list will be empty if the metadata level is 'none'
        or 'level1'.
    performer_radius: float
        The radius of the performer, in meters.
    performer_reach: float
        The max reach of the performer, in meters.
    physics_frames_per_second : float
        The frames per second of the physics engine
    position : dict
        The "x", "y", and "z" coordinates for your global position.
        Will be set to 'None' if using a metadata level below the
        'oracle' level.
    resolved_object : string
        The object that was selected based on objectImageCoords
    resolved_receptacle_object_id : string
        The receptacle that was selected based on receptacleObjectImageCoords
    return_status : string
        The return status from your last action. See
        :mod:`Action <machine_common_sense.Action>`.
    reward : integer
        Reward is 1 on successful completion of a task, 0 otherwise.
    room_dimensions : dict
        The "x", "y", and "z" dimensions of the current scene.
        Will be set to 'None' if using a metadata level below the
        'oracle' level.
    rotation : float
        Your current rotation angle in degrees. Will be set to 'None'
        if using a metadata level below the 'oracle' level.
    segmentation_colors : list of dicts
        The colors for all objects in the instance segmentation images
        (in `object_mask_list`), each represented as a dict containing an
        "objectId" string property and "r", "g", and "b" int properties for the
        corresponding red, green, and blue values. The ceiling has an objectId
        of "ceiling"; exterior room walls have objectIds of "wall_back",
        "wall_front", "wall_left", and "wall_right"; floor sections have
        objectIds starting with "floor " and then the texture name (since
        different areas of the floor can have different textures); holes have
        objectIds of "hole"; hole walls have objectIds of "hole wall"; and
        lava areas have objectIds of "lava".
        Will be empty if using a metadata level below the 'oracle' level.
    step_number : integer
        The step number of your last action, recorded since you started the
        current scene.
    steps_in_lava : integer
        The number of steps the agent has touched lava
    structural_object_list : list of ObjectMetadata objects
        The list of metadata for all the visible structural objects (like
        walls, occluders, and ramps) in the scene. This list will be empty
        if using a metadata level below the 'oracle' level.
        Occluders are composed of two separate objects,
        the "wall" and the "pole", with corresponding object IDs
        (occluder_wall_<uuid> and occluder_pole_<uuid>), and ramps are
        composed of between one and three objects (depending on the type
        of ramp), with corresponding object IDs.
    triggered_by_sequence_incorrect : bool
        If the the sequence to trigger a placer holding the target is incorrect
    """

    def __init__(
        self,
        action_list=None,
        camera_aspect_ratio=None,
        camera_clipping_planes=None,
        camera_field_of_view=0.0,
        camera_height=0.0,
        depth_map_list=None,
        goal=None,
        habituation_trial=None,
        haptic_feedback=None,
        head_tilt=0.0,
        holes=None,
        image_list=None,
        lava=None,
        object_list=None,
        object_mask_list=None,
        performer_radius=0.0,
        performer_reach=0.0,
        physics_frames_per_second=0,
        position=None,
        resolved_object='',
        resolved_receptacle='',
        return_status=ReturnStatus.UNDEFINED.value,
        reward=0,
        room_dimensions=None,
        rotation=0.0,
        segmentation_colors=None,
        step_number=0,
        steps_on_lava=0,
        structural_object_list=None,
        triggered_by_sequence_incorrect=False
    ):
        self.action_list = [] if action_list is None else action_list
        self.camera_aspect_ratio = (
            0.0, 0.0) if camera_aspect_ratio is None else camera_aspect_ratio
        self.camera_clipping_planes = (
            (0.0, 0.0)
            if camera_clipping_planes is None
            else camera_clipping_planes
        )
        self.camera_field_of_view = camera_field_of_view
        self.camera_height = camera_height
        self.depth_map_list = (
            [] if depth_map_list is None else depth_map_list
        )
        self.goal = GoalMetadata() if goal is None else goal
        self.habituation_trial = habituation_trial
        self.haptic_feedback = (
            {} if haptic_feedback is None else haptic_feedback
        )
        self.head_tilt = head_tilt
        self.holes = [] if holes is None else holes
        self.image_list = [] if image_list is None else image_list
        self.lava = [] if lava is None else lava
        self.object_list = [] if object_list is None else object_list
        self.object_mask_list = (
            [] if object_mask_list is None else object_mask_list
        )
        self.performer_radius = performer_radius
        self.performer_reach = performer_reach
        self.physics_frames_per_second = physics_frames_per_second
        self.position = {} if position is None else position
        self.resolved_object = resolved_object
        self.resolved_receptacle = resolved_receptacle
        self.return_status = return_status
        self.reward = reward
        self.room_dimensions = (
            {} if room_dimensions is None else room_dimensions
        )
        self.rotation = rotation
        self.segmentation_colors = (
            [] if segmentation_colors is None else segmentation_colors
        )
        self.step_number = step_number
        self.steps_on_lava = steps_on_lava
        self.structural_object_list = [
        ] if structural_object_list is None else structural_object_list
        self.triggered_by_sequence_incorrect = triggered_by_sequence_incorrect

    def __str__(self):
        return Stringifier.class_to_str(self)

    def check_list_none(self, obj_list):
        if obj_list is None:
            return None
        else:
            return {obj.uuid: dict(obj) for obj in obj_list}

    def copy_without_depth_or_images(self):
        """Return a deep copy of this StepMetadata with default depth_map_list,
        image_list, and object_mask_list properties."""
        step_metadata_copy = StepMetadata()
        # This class's __iter__ function will ignore specific properties.
        for key, _ in self:
            setattr(step_metadata_copy, key, copy.deepcopy(getattr(self, key)))
        return step_metadata_copy

    # Allows converting the class to a dictionary, along with allowing
    #   certain fields to be left out of output file
    def __iter__(self):
        yield 'action_list', self.action_list
        yield 'camera_aspect_ratio', self.camera_aspect_ratio
        yield 'camera_clipping_planes', self.camera_clipping_planes
        yield 'camera_field_of_view', self.camera_field_of_view
        yield 'camera_height', self.camera_height
        # Intentionally no depth_map_list
        yield 'goal', dict(self.goal)
        yield 'habituation_trial', self.habituation_trial
        yield 'haptic_feedback', self.haptic_feedback
        yield 'head_tilt', self.head_tilt
        yield 'holes', self.head_tilt
        # Intentionally no image_list
        yield 'lava', self.head_tilt
        yield 'object_list', self.check_list_none(self.object_list)
        # Intentionally no object_mask_list
        yield 'performer_radius', self.performer_radius
        yield 'performer_reach', self.performer_reach
        yield 'physics_frames_per_second', self.physics_frames_per_second
        yield 'position', self.position
        yield 'resolved_object', self.resolved_object
        yield 'resolved_receptacle', self.resolved_receptacle
        yield 'return_status', self.return_status
        yield 'room_dimensions', self.room_dimensions
        yield 'reward', self.reward
        yield 'rotation', self.rotation
        yield 'segmentation_colors', self.segmentation_colors
        yield 'step_number', self.step_number
        yield 'steps_on_lava', self.steps_on_lava
        yield 'structural_object_list', self.check_list_none(
            self.structural_object_list)
        yield 'triggered_by_sequence_incorrect', \
            self.triggered_by_sequence_incorrect



import logging

import numpy

logger = logging.getLogger(__name__)


class Stringifier:
    """
    Defines functions to turn objects into strings for debugging and
    human readable output.  It is not intended to be reconstructed
    which is why this is seperate from serialization
    """

    NUMBER_OF_DECIMALS = 6
    NUMBER_OF_SPACES = 4

    @staticmethod
    def class_to_str(input_class, depth=0):
        """
        Transforms the given class into a string.

        Parameters
        ----------
        input_value
            The input class.
        depth : int, optional
            The indent depth (default 0).

        Returns
        -------
        string
        """
        this_indent = " " * Stringifier.NUMBER_OF_SPACES * depth
        next_indent = " " * Stringifier.NUMBER_OF_SPACES * (depth + 1)
        props = {
            prop_key: prop_value
            for prop_key, prop_value in vars(input_class).items()
            if not prop_key.startswith('_') or callable(prop_value)
        }
        text_list = [next_indent +
                     "\"" +
                     prop_key +
                     "\": " +
                     Stringifier.value_to_str(
                         prop_value,
                         depth +
                         1) for prop_key, prop_value in props.items()]
        return "{}" if not text_list else "{\n" + \
            (",\n").join(text_list) + "\n" + this_indent + "}"

    @staticmethod
    def generate_pretty_object_output(object_list):
        """
        Transforms the given list of ObjectMetadata objects into a
        list of strings.

        Parameters
        ----------
        object_list : list of ObjectMetadata objects
            The input list.

        Returns
        -------
        list of strings
        """
        # TODO What else should we show here?
        titles = [
            "OBJECT ID",
            "SHAPE",
            "COLORS",
            "HELD",
            "VISIBLE",
            "STATE",
            "POSITION (WORLD)",
            "DISTANCE (WORLD)",
            "DIRECTION (WORLD)",
            "DIMENSIONS (WORLD)"
        ]
        rows = [titles] + [
            [
                metadata.uuid,
                metadata.shape,
                ", ".join(metadata.texture_color_list)
                if (metadata.texture_color_list is not None)
                else metadata.texture_color_list,
                metadata.held,
                metadata.visible,
                ", ".join(metadata.state_list)
                if (metadata.state_list is not None)
                else metadata.state_list,
                Stringifier.vector_to_string(metadata.position),
                metadata.distance_in_world,
                Stringifier.vector_to_string(metadata.direction),
                (
                    (
                        "[" +
                        ", ".join(
                            Stringifier.vector_to_string(corner)
                            for corner in metadata.dimensions
                        )
                    ) +
                    "]"
                )
                if metadata.dimensions
                else None,
            ]
            for metadata in object_list
        ]

        widths = [max(len(str(row[i])) for row in rows)
                  for i in range(len(titles))]
        return [
            "  ".join(str(row[i]).ljust(widths[i]) for i in range(len(row)))
            for row in rows
        ]

    @staticmethod
    def value_to_str(input_value, depth=0):
        """
        Transforms the given value into a string.

        Parameters
        ----------
        input_value
            The input value.
        depth : int, optional
            The indent depth (default 0).

        Returns
        -------
        string
        """
        this_indent = " " * Stringifier.NUMBER_OF_SPACES * depth
        next_indent = " " * Stringifier.NUMBER_OF_SPACES * (depth + 1)
        if input_value is None:
            return "null"
        if isinstance(input_value, dict):
            text_list = [
                next_indent + "\"" + str(dict_key) + "\": " +
                Stringifier.value_to_str(dict_value, depth + 1)
                for dict_key, dict_value in input_value.items()
            ]
            return "{}" if not text_list else "{\n" + \
                (",\n").join(text_list) + "\n" + this_indent + "}"
        if isinstance(input_value, (list, tuple, numpy.ndarray)):
            input_value_as_list = list(input_value)
            condense = not any(
                isinstance(list_item, (dict, list, tuple, numpy.ndarray))
                for list_item in input_value_as_list
            )

            if condense:
                # To condense the list output, remove all of the whitespace.
                text_list = [
                    Stringifier.value_to_str(list_item, 0)
                    for list_item in input_value_as_list
                ]
                return "[" + (",").join(text_list) + "]"
            # Else the list output will separate list elements with newlines.
            text_list = [
                next_indent + Stringifier.value_to_str(list_item, depth + 1)
                for list_item in input_value_as_list
            ]
            return "[]" if not text_list else "[\n" + \
                (",\n").join(text_list) + "\n" + this_indent + "]"
        if isinstance(input_value, bool):
            return "true" if input_value else "false"
        if isinstance(input_value, (float, numpy.float32, numpy.float64)):
            return str(round(input_value, Stringifier.NUMBER_OF_DECIMALS))
        if isinstance(input_value, str):
            return "\"" + input_value.replace("\"", "\\\"") + "\""
        return str(input_value).replace("\n", "\n" + this_indent)

    @staticmethod
    def vector_to_string(vector):
        """
        Transforms the given vector into a string.

        Parameters
        ----------
        vector : dict
            The input vector.

        Returns
        -------
        string
        """
        return (
            (
                '(' + str(round(vector['x'], Stringifier.NUMBER_OF_DECIMALS)) +
                ',' + str(round(vector['y'], Stringifier.NUMBER_OF_DECIMALS)) +
                ',' + str(round(vector['z'],
                                Stringifier.NUMBER_OF_DECIMALS)) + ')'
            )
            if vector is not None and 'x' in vector and 'y' in vector and
            'z' in vector else 'None'
        )



from __future__ import annotations

import math
from dataclasses import dataclass
from typing import Dict, List, Tuple

import ai2thor.server
import colour
import cv2
import numpy as np
import PIL.Image
import skimage.draw
from shapely import affinity, geometry

from machine_common_sense.config_manager import SceneConfiguration, Vector3d


@dataclass
class SceneCoord():
    '''Unity scenes have floor dimensions of x and z'''
    x: float
    y: float
    z: float

    def __sub__(self, coord: SceneCoord):
        return SceneCoord(self.x - coord.x, self.y - coord.y, self.z - coord.z)

    def __add__(self, coord: SceneCoord):
        return SceneCoord(self.x + coord.x, self.y + coord.y, self.z + coord.z)

    def __or__(self, coord: SceneCoord):
        '''Midpoint between two SceneCoords'''
        return SceneCoord(
            (self.x + coord.x) / 2,
            (self.y + coord.y) / 2,
            (self.z + coord.z) / 2
        )


@dataclass
class ImageCoord():
    '''Image xy coordinates where y corresponds to z in Unity'''
    x: int
    y: int


@dataclass
class XZHeading():
    '''Unity directional vector'''
    x: float
    z: float


@dataclass
class Robot():
    '''AI robot/performer location and rotation in Unity'''
    x: float
    y: float
    z: float
    rotation_y: float


@dataclass
class Arrow():
    floor: List[SceneCoord]
    peak: SceneCoord


@dataclass
class SceneBounds():
    points: List[SceneCoord]
    rotation: Dict = None

    def __post_init__(self):
        # TODO If this object has any X/Z rotation, rotate the corners too.
        # Safe to assume the bounds will always be in a consistent order.
        # See BaseFPSAgentController.WorldCoordinatesOfBoundingBox (Unity)
        assert len(self.points) == 8

    def bottom_front_right_corner(self) -> SceneCoord:
        return self.points[0]

    def bottom_front_left_corner(self) -> SceneCoord:
        return self.points[1]

    def bottom_back_left_corner(self) -> SceneCoord:
        return self.points[2]

    def bottom_back_right_corner(self) -> SceneCoord:
        return self.points[3]

    def top_front_right_corner(self) -> SceneCoord:
        return self.points[4]

    def top_front_left_corner(self) -> SceneCoord:
        return self.points[5]

    def top_back_left_corner(self) -> SceneCoord:
        return self.points[6]

    def top_back_right_corner(self) -> SceneCoord:
        return self.points[7]


@dataclass
class SceneAsset():
    held: bool
    visible: bool
    uuid: str
    color: str
    bounds: SceneBounds


@dataclass
class Ramp(SceneAsset):

    @property
    def arrow(self) -> Arrow:
        peak_mid = self._peak_midpoint()
        floor_pts = self._floor_points()
        return Arrow(floor=floor_pts, peak=peak_mid)

    def _peak_midpoint(self) -> SceneCoord:
        '''The midpoint of the highest ramp side'''
        # Assumes ramps are always wedges (triangles).
        peak_pts = [
            self.bounds.top_front_right_corner(),
            self.bounds.top_front_left_corner()
        ]
        peak_scene_pts = [
            SceneCoord(x=pt.x, y=pt.y, z=pt.z) for pt in peak_pts
        ]
        return peak_scene_pts[0] | peak_scene_pts[1]

    def _floor_points(self) -> List[SceneCoord]:
        '''The points of the ramp side nearest the floor'''
        # Assumes ramps are always wedges (triangles).
        floor_points = [
            self.bounds.bottom_back_right_corner(),
            self.bounds.bottom_back_left_corner()
        ]
        return [SceneCoord(pt.x, pt.y, pt.z) for pt in floor_points]


@dataclass
class Texture():
    material: str
    positions: list


class TopDownPlotter():

    DEFAULT_COLOR = colour.COLOR_NAME_TO_RGB['white']
    BACKGROUND_COLOR = colour.COLOR_NAME_TO_RGB['black']
    BORDER_COLOR = colour.COLOR_NAME_TO_RGB['white']
    CENTER_COLOR = colour.COLOR_NAME_TO_RGB['gray']
    GRID_COLOR = colour.COLOR_NAME_TO_RGB['darkslategray']
    ROBOT_COLOR = colour.COLOR_NAME_TO_RGB['blue']
    DEFAULT_GOAL_COLOR = colour.COLOR_NAME_TO_RGB['lime']

    OBJECT_INSIDE_PERCENTAGE = 0.5
    HEADING_LENGTH = 0.6
    PLOT_IMAGE_SIZE = 512
    UNIT_CELL_WIDTH = 1

    FONT = cv2.FONT_HERSHEY_COMPLEX
    FONT_SCALE = 0.4
    FONT_THICKNESS = 1
    WALL_BUFFER = 50

    def __init__(self, team: str, scene_config: SceneConfiguration) -> None:
        self._scene_name = scene_config.name.replace('json', '')
        if '/' in self._scene_name:
            self._scene_name = self._scene_name.rsplit('/', 1)[1]

        self._room_size = (
            # Room is automatically expanded in intuitive physics scenes.
            Vector3d(x=14, y=10, z=10) if scene_config.intuitive_physics
            else scene_config.room_dimensions
        )
        self._team = team
        # create the room once as it is computationally expensive
        self.base_room_img = self._initialize_plot(scene_config)

    def _initialize_plot(self, scene_config: SceneConfiguration) -> np.ndarray:
        '''Create the initial plot with grid lines'''
        img = np.zeros(
            (self.PLOT_IMAGE_SIZE, self.PLOT_IMAGE_SIZE, 3),
            dtype=np.int8)

        room = SceneCoord(
            x=self._room_size.x,
            y=self._room_size.y,
            z=self._room_size.z
        )
        largest_dimension = max(room.x, room.z)

        x_dim, y_dim, _ = img.shape
        self.image_center = ImageCoord(
            x=int(x_dim / 2) - 1,  # Ex. 512/2-1 = 255
            y=int(y_dim / 2) - 1
        )

        # determine the scale for each xy dimension
        self.scale = ImageCoord(
            x=(x_dim - self.WALL_BUFFER) / largest_dimension,
            y=(y_dim - self.WALL_BUFFER) / largest_dimension
        )

        # get the buffer for both x and y especially for non-square rooms
        buffer = ImageCoord(
            x=(x_dim - self.scale.x * room.x) / 2,
            y=(y_dim - self.scale.y * room.z) / 2
        )

        img = self._draw_grid_lines(img, buffer)

        img = self._draw_holes(
            img,
            scene_config.holes,
            scene_config.room_dimensions)
        img = self._draw_lava(
            img,
            scene_config.retrieve_lava(),
            scene_config.room_dimensions)

        return img

    def plot(self, scene_event: ai2thor.server.Event,
             step_number: int, goal_ids: list = None) -> PIL.Image.Image:
        '''Create a plot of the room, objects and robot'''
        plt_img = self.base_room_img.copy()

        plt_objects = self._find_plottable_objects(scene_event)
        plt_img = self._draw_objects(plt_img,
                                     plt_objects,
                                     goal_ids)

        agent_metadata = scene_event.metadata.get('agent')
        plt_img = self._draw_robot(
            plt_img,
            agent_metadata)

        plt_img = self._draw_step_number(plt_img, step_number)
        return self._export_plot(img=plt_img)

    def _convert_to_image_coords(self, scene_pt: SceneCoord) -> ImageCoord:
        '''converts scene xz coordinate point to an xy image coordinate'''
        return ImageCoord(
            x=int(self.image_center.x + scene_pt.x * self.scale.x),
            y=int(self.image_center.y - scene_pt.z * self.scale.y)
        )

    def _convert_points(self, points: list) -> tuple:
        # convert list of tuples to list of rows and list of columns
        cs, rs = map(list, zip(*points))
        # convert room coordinates to image coordinates
        rs = list(map(lambda r: self.image_center.y - r * self.scale.y, rs))
        cs = list(map(lambda c: self.image_center.x + c * self.scale.x, cs))
        return cs, rs

    def _draw_step_number(self, img: np.ndarray,
                          step_number: int) -> np.ndarray:
        '''Write the step number on the plot image'''
        (label_width, label_height), baseline = cv2.getTextSize(
            str(step_number), self.FONT, self.FONT_SCALE, self.FONT_THICKNESS)
        text_draw_buffer = 5
        cv2.putText(
            img=img,
            text=str(step_number),
            org=(img.shape[0] - label_width - text_draw_buffer,
                 label_height + baseline),
            fontFace=cv2.FONT_HERSHEY_SIMPLEX,
            fontScale=self.FONT_SCALE,
            color=self.BORDER_COLOR,
            lineType=self.FONT_THICKNESS)
        return img

    def _draw_grid_lines(self, img: np.ndarray,
                         buffer: ImageCoord) -> np.ndarray:
        '''Draw room coordinate grid aligned with Unity units'''
        img = self._draw_vertical_grid_lines(img, buffer.y)
        img = self._draw_horizontal_grid_lines(img, buffer.x)
        return self._draw_room_border(img, buffer)

    def _draw_vertical_grid_lines(
            self, img: np.ndarray, buffer_y: int) -> np.ndarray:
        '''Draw vertical lines at each x-coordinate'''
        x_units = math.floor(self._room_size.x / 2)
        # draw vertical grid lines starting at the center and working out
        for x in range(x_units + 1):
            # positive x values
            img = self._draw_vertical_line(img, buffer_y, x)
            # negative x values
            img = self._draw_vertical_line(img, buffer_y, -x)
        return img

    def _draw_vertical_line(self, img: np.ndarray,
                            buffer_y: int, x_coord: int) -> np.ndarray:
        '''Draw a vertical grid line corresponding to the x-coordinate'''
        img_x = int(self.image_center.x + x_coord * self.scale.x)
        rr, cc = skimage.draw.line(
            c0=img_x,
            r0=int(buffer_y),
            c1=img_x,
            r1=img.shape[0] - 1 - int(buffer_y))
        img[rr, cc] = self.CENTER_COLOR if x_coord == 0 else self.GRID_COLOR
        return img

    def _draw_horizontal_grid_lines(
            self, img: np.ndarray, buffer_x: int) -> np.ndarray:
        '''Draw horizontal lines at each y-coordinate'''
        z_units = math.floor(self._room_size.z / 2)
        # draw horizontal grid lines starting at the center and working out
        for y in range(z_units + 1):
            img = self._draw_horizontal_line(img, buffer_x, y)
            img = self._draw_horizontal_line(img, buffer_x, -y)
        return img

    def _draw_horizontal_line(self, img: np.ndarray,
                              buffer_x: int, y_coord: int) -> np.ndarray:
        '''Draw a horizontal grid line corresponding to the y-coordinate'''
        img_y = int(self.image_center.y + y_coord * self.scale.y)
        rr, cc = skimage.draw.line(
            c0=int(buffer_x),
            r0=img_y,
            c1=img.shape[1] - 1 - int(buffer_x),
            r1=img_y)
        img[rr, cc] = self.CENTER_COLOR if y_coord == 0 else self.GRID_COLOR
        return img

    def _draw_room_border(self, img: np.ndarray,
                          buffer: ImageCoord) -> np.ndarray:
        '''Outline the room border'''
        x_dim, z_dim, _ = img.shape
        rr, cc = skimage.draw.rectangle_perimeter(
            start=(buffer.y, buffer.x),
            end=(z_dim - buffer.y - 1, x_dim - buffer.x - 1),
            shape=img.shape[:2],
            clip=True)
        img[rr, cc] = self.BORDER_COLOR
        return img

    def _out_of_bounds(
            self, pos: SceneCoord, room_dim: Vector3d) -> bool:
        x_odd_reducer = 0 if room_dim.x % 2 == 0 else 1
        z_odd_reducer = 0 if room_dim.z % 2 == 0 else 1
        return (abs(pos.x) > (room_dim.x - x_odd_reducer) /
                2 or abs(pos.z) > (room_dim.z - z_odd_reducer) / 2)

    def _even_room_dim_edge_size_reducer(
            self,
            pos: SceneCoord,
            room_dim: Vector3d = Vector3d(x=10, y=0, z=10)
    ) -> Tuple(SceneCoord):

        size_reducer_left = SceneCoord(0, 0, 0)
        size_reducer_right = SceneCoord(0, 0, 0)
        if pos.x == -(room_dim.x / 2):
            size_reducer_left.x = 0.5
        if pos.z == -(room_dim.z / 2):
            size_reducer_left.z = 0.5
        if pos.x == (room_dim.x / 2):
            size_reducer_right.x = -0.5
        if pos.z == (room_dim.z / 2):
            size_reducer_right.z = -0.5

        return size_reducer_left, size_reducer_right

    def _even_x_odd_z_room_dim_edge_size_reducer(
            self,
            pos: SceneCoord,
            room_dim: Vector3d = Vector3d(x=10, y=0, z=10)
    ) -> Tuple(SceneCoord):

        size_reducer_left = SceneCoord(0, 0, 0)
        size_reducer_right = SceneCoord(0, 0, 0)
        if pos.x == -(room_dim.x / 2):
            size_reducer_left.x = 0.5
        if pos.x == (room_dim.x / 2):
            size_reducer_right.x = -0.5

        return size_reducer_left, size_reducer_right

    def _odd_x_even_z_room_dim_edge_size_reducer(
            self,
            pos: SceneCoord,
            room_dim: Vector3d = Vector3d(x=10, y=0, z=10)
    ) -> Tuple(SceneCoord):

        size_reducer_left = SceneCoord(0, 0, 0)
        size_reducer_right = SceneCoord(0, 0, 0)
        if pos.z == -(room_dim.z / 2):
            size_reducer_left.z = 0.5
        if pos.z == (room_dim.z / 2):
            size_reducer_right.z = -0.5

        return size_reducer_left, size_reducer_right

    def _plotter_hole_texture_size_reducer(
            self,
            pos: SceneCoord,
            room_dim: Vector3d = Vector3d(x=10, y=0, z=10)
    ) -> Tuple(SceneCoord):

        x_is_even = room_dim.x % 2 == 0
        z_is_even = room_dim.z % 2 == 0

        if self._out_of_bounds(pos, room_dim):
            return None, None
        if x_is_even and z_is_even:
            return (self._even_room_dim_edge_size_reducer(pos, room_dim))
        if x_is_even and not z_is_even:
            return self._even_x_odd_z_room_dim_edge_size_reducer(pos, room_dim)
        return self._odd_x_even_z_room_dim_edge_size_reducer(pos, room_dim)

    def _draw_lava(self, img: np.ndarray, lava: List,
                   room_dim: Vector3d = Vector3d(x=10, y=0, z=10)
                   ) -> np.ndarray:
        if lava is None:
            return img

        for area in lava:
            area_pos_ul = SceneCoord(x=area[0], y=0, z=area[1])
            area_pos_lr = SceneCoord(x=area[2], y=0, z=area[3])
            midpoint = area_pos_ul | area_pos_lr
            size_reducer_left, size_reducer_right = (
                self._plotter_hole_texture_size_reducer(midpoint, room_dim)
            )
            if size_reducer_left is None:
                continue
            area_pos_ul = area_pos_ul + size_reducer_left
            area_pos_lr = area_pos_lr + size_reducer_right
            area_img_pos_ul = self._convert_to_image_coords(area_pos_ul)
            area_img_pos_lr = self._convert_to_image_coords(area_pos_lr)

            img = self._draw_lava_area(
                img=img,
                upper_left=area_img_pos_ul,
                lower_right=area_img_pos_lr,
                texture_color=colour.COLOR_NAME_TO_RGB['red'])
            img = self._draw_perimeter(
                img,
                area_img_pos_ul,
                area_img_pos_lr,
                self.BACKGROUND_COLOR)
            img = self._draw_x(
                img,
                area_img_pos_ul,
                area_img_pos_lr,
                self.BACKGROUND_COLOR)
        return img

    def _draw_lava_area(
        self,
        img: np.ndarray,
        upper_left: ImageCoord,
        lower_right: ImageCoord,
        texture_color: colour.C_RGB
    ) -> np.ndarray:
        rr, cc = skimage.draw.rectangle(
            start=(upper_left.y, upper_left.x),
            end=(lower_right.y, lower_right.x),
            shape=img.shape[:2])
        img[rr, cc] = texture_color
        return img

    def _draw_holes(self, img: np.ndarray, holes: List,
                    room_dim: Vector3d = Vector3d(x=10, y=0, z=10
                                                  )) -> np.ndarray:
        '''Draw a box with an X to illustrate a floor hole'''
        if holes is None:
            return img

        for hole in holes:
            hole_center = SceneCoord(x=hole.x, y=0, z=hole.z)

            size_reducer_left, size_reducer_right = (
                self._plotter_hole_texture_size_reducer(
                    hole, room_dim))
            if size_reducer_left is None:
                continue
            half_cell_pos = SceneCoord(
                self.UNIT_CELL_WIDTH / 2, 0, self.UNIT_CELL_WIDTH / 2)
            # calculate scene corners
            hole_upper_left: SceneCoord = hole_center - half_cell_pos + \
                size_reducer_left
            hole_lower_right: SceneCoord = hole_center + half_cell_pos + \
                size_reducer_right

            # convert scene corners to image coordinates
            hole_img_upper_left: ImageCoord = self._convert_to_image_coords(
                hole_upper_left)
            hole_img_lower_right: ImageCoord = self._convert_to_image_coords(
                hole_lower_right)

            img = self._draw_perimeter(
                img,
                hole_img_upper_left,
                hole_img_lower_right,
                self.DEFAULT_COLOR)
            img = self._draw_x(
                img,
                hole_img_upper_left,
                hole_img_lower_right,
                self.DEFAULT_COLOR)
        return img

    def _draw_perimeter(
            self,
            img: np.ndarray,
            upper_left: ImageCoord,
            lower_right: ImageCoord,
            color: Tuple) -> np.ndarray:
        '''Outline the floor hole'''
        rr, cc = skimage.draw.rectangle_perimeter(
            start=(upper_left.y - 1, upper_left.x + 1),
            end=(lower_right.y + 1, lower_right.x - 1),
            shape=img.shape[:2],
            clip=True)
        img[rr, cc] = color
        return img

    def _draw_x(
            self,
            img: np.ndarray,
            upper_left: ImageCoord,
            lower_right: ImageCoord,
            color: Tuple) -> np.ndarray:
        rr, cc = skimage.draw.line(
            r0=upper_left.y,
            c0=upper_left.x,
            r1=lower_right.y,
            c1=lower_right.x)
        img[rr, cc] = color

        rr, cc = skimage.draw.line(
            r0=lower_right.y,
            c0=upper_left.x,
            r1=upper_left.y,
            c1=lower_right.x)
        img[rr, cc] = color
        return img

    def _export_plot(self, img: np.ndarray) -> PIL.Image.Image:
        '''Export the plot to a PIL Image'''
        return PIL.Image.fromarray(img, "RGB")

    def _draw_robot(self, img: np.ndarray, robot_metadata: Dict) -> np.ndarray:
        '''Plot the robot triangle'''
        if robot_metadata is None:
            return img
        robot = self._create_robot(robot_metadata)
        img = self._draw_robot_position(img, robot)
        return img

    def _create_robot(self, robot_metadata: Dict) -> Robot:
        '''Extract robot position and rotation information from the metadata'''
        position = robot_metadata.get('position')
        if position is not None:
            x = position.get('x')
            y = position.get('y')
            z = position.get('z')
        else:
            x = 0.0
            y = 0.0
            z = 0.0
        rotation = robot_metadata.get('rotation')
        rotation_y = rotation.get('y') if rotation is not None else 0.0
        return Robot(x, y, z, 360 - rotation_y)

    def _calculate_heading(self, rotation_angle: float,
                           heading_length: float) -> XZHeading:
        '''Calculate XZ heading vector from the rotation angle'''
        s = math.sin(math.radians(rotation_angle))
        c = math.cos(math.radians(rotation_angle))
        vec_x = heading_length * s * -1
        vec_z = heading_length * c
        return XZHeading(vec_x, vec_z)

    def _draw_robot_position(self, img: np.ndarray,
                             robot: Robot) -> np.ndarray:
        '''Draw the robot's scene XZ position as a triangle in the plot'''
        opposite_color = self._find_opposite_color(self.ROBOT_COLOR)
        half_length = self.HEADING_LENGTH / 2.0
        # Create the triangle points.
        front = self._calculate_heading(robot.rotation_y, self.HEADING_LENGTH)
        right = self._calculate_heading(120 + robot.rotation_y, half_length)
        left = self._calculate_heading(240 + robot.rotation_y, half_length)
        triangle = [(front.x, front.z), (right.x, right.z), (left.x, left.z)]
        # Create the triangle polygon.
        outer = geometry.MultiPoint(triangle).convex_hull
        # Move the triangle from the origin to the correct position.
        outer = affinity.translate(outer, robot.x, robot.z)
        # Draw the robot as two triangles: first a large outer triangle of the
        # opposite color as a border; and a small inner triangle as the center.
        percent = self.OBJECT_INSIDE_PERCENTAGE
        inner = affinity.scale(outer, xfact=percent, yfact=percent)
        for polygon, color in [
            (outer, opposite_color),
            (inner, self.ROBOT_COLOR)
        ]:
            points = list(polygon.exterior.coords)
            cs, rs = self._convert_points(points)
            rr, cc = skimage.draw.polygon(rs, cs, shape=img.shape[:2])
            img[rr, cc] = color
        return img

    def _find_plottable_objects(
            self, scene_event: ai2thor.server.Event) -> List:
        '''Find plottable objects from the scene data.

        Plottable objects include normal scene objects as well as
        occluder and wall structural objects.
        '''
        structural_objects = scene_event.metadata.get('structuralObjects', [])
        filtered_structural_objects = [
            obj for obj in structural_objects
            if not obj.get('objectId', '').startswith('ceiling') and not
            obj.get('objectId', '').startswith('floor')
        ]
        objects = scene_event.metadata.get('objects', [])
        combined_objects = filtered_structural_objects + objects
        # the combined list needs to be sorted by y height
        # so that the lowest objects are drawn first
        combined_objects = sorted(
            combined_objects,
            key=lambda obj: obj['objectBounds']['objectBoundsCorners'][0]['y'])
        return combined_objects

    def _draw_objects(self, img: np.ndarray, objects: Dict,
                      goal_ids: list = None) -> np.ndarray:
        '''Plot the object bounds for each object in the scene'''
        for o in objects:
            obj = self._create_asset(o)
            if obj.bounds is not None:
                if goal_ids is not None and o['objectId'] in goal_ids:
                    img = self._draw_goal(img, obj)
                else:
                    img = self._draw_object(img, obj)
        return img

    def _create_asset(self, object_metadata: Dict) -> SceneAsset:
        '''Create the scene object from its metadata'''
        colors = object_metadata.get('colorsFromMaterials', [])
        if len(colors):
            color = self._convert_color(colors[0])
        else:
            color = self._convert_color('')

        obj_bounds = object_metadata.get('objectBounds')
        corners = None if obj_bounds is None else obj_bounds.get(
            'objectBoundsCorners')
        bounds = None if corners is None else [SceneCoord(
            x=corner['x'], y=corner['y'], z=corner['z']) for corner in corners]

        return SceneAsset(
            held=object_metadata.get('isPickedUp'),
            visible=object_metadata.get('visibleInCamera'),
            uuid=object_metadata.get('objectId'),
            color=color,
            bounds=(
                SceneBounds(bounds, object_metadata.get('rotation'))
                if bounds else None
            )
        )

    def _convert_color(self, color: str) -> str:
        '''Convert color string to string'''
        if not color:
            color = 'ivory'
        # Don't use the same color as the background (black).
        if color == 'black':
            color = 'ivory'
        # Rose is apparently not a valid web color.
        if color == 'rose':
            color = 'deeppink'
        # Azure as a web color is very different from the normal azure.
        if color == 'azure':
            color = 'deepskyblue'
        return color

    def _draw_object(self, img: np.ndarray,
                     obj: SceneAsset, is_goal: bool = False) -> np.ndarray:
        '''Draw the scene object'''

        obj_pts = [(pt.x, pt.z) for pt in obj.bounds.points]
        polygon = geometry.MultiPoint(obj_pts).convex_hull
        # Shrink the polygon for the goal object, since it will be drawn on top
        # of the correctly-sized polygon from _draw_goal.
        if is_goal:
            percent = self.OBJECT_INSIDE_PERCENTAGE
            polygon = affinity.scale(polygon, xfact=percent, yfact=percent)
        pts = polygon.exterior.coords

        cs, rs = self._convert_points(pts)

        # use dictionary get method for color retrieval
        # if no match for that color string, then resort to the default color
        clr = colour.COLOR_NAME_TO_RGB.get(
            obj.color.lower(), self.DEFAULT_COLOR)

        # draw filled polygon if visible to the robot
        if is_goal or obj.visible:
            rr, cc = skimage.draw.polygon(
                rs,
                cs,
                shape=img.shape[:2])
            # Draw a small black polygon for a hidden goal object, since it
            # already has a border around it.
            if not obj.visible:
                clr = self.BACKGROUND_COLOR
        else:
            rr, cc = skimage.draw.polygon_perimeter(
                rs,
                cs,
                shape=img.shape[:2])
        img[rr, cc] = clr

        # using ramp string prefix assumpation to make ramp determination
        # might be better to have an attribute to leverage
        if(obj.uuid.startswith('ramp')):
            self._draw_ramp_arrow(img, obj, clr)

        return img

    def _draw_ramp_arrow(self, img: np.ndarray, obj: SceneAsset,
                         ramp_color: Tuple) -> np.ndarray:
        # convert scene object to a ramp
        ramp = obj
        ramp.__class__ = Ramp
        return self._draw_arrow(img, ramp, ramp_color)

    def _draw_arrow(self, img: np.ndarray, ramp: Ramp,
                    ramp_color: Tuple) -> np.ndarray:
        '''draw lines from the arrow floor points to the peak
        to illustrate the ramp
        '''
        peak_pt = self._convert_to_image_coords(ramp.arrow.peak)
        for floor_pt in ramp.arrow.floor:
            img_pt = self._convert_to_image_coords(floor_pt)
            rr, cc = skimage.draw.line(
                r0=peak_pt.y,
                c0=peak_pt.x,
                r1=img_pt.y,
                c1=img_pt.x)
            arrow_color = (
                self.BACKGROUND_COLOR if ramp_color != self.BACKGROUND_COLOR
                else self.DEFAULT_COLOR
            )
            img[rr, cc] = arrow_color if ramp.visible else ramp_color
        return img

    def _draw_goal(self, img: np.ndarray,
                   obj: SceneAsset) -> np.ndarray:
        '''Draw the goal object of the scene'''
        obj_pts = [(pt.x, pt.z) for pt in obj.bounds.points]
        polygon = geometry.MultiPoint(
            obj_pts).convex_hull
        pts = polygon.exterior.coords

        color = colour.COLOR_NAME_TO_RGB.get(
            obj.color.lower(),
            self.DEFAULT_COLOR
        )
        opposite_color = self._find_opposite_color(color)
        # If the object's color is greyscale, use a default instead.
        if color[0] == color[1] == color[2]:
            opposite_color = self.DEFAULT_GOAL_COLOR

        cs, rs = self._convert_points(pts)

        # Draw a correctly-sized polygon with the opposite of the goal's color.
        # This will create a thick border around the 2nd polygon (see below).
        rr, cc = skimage.draw.polygon(rs, cs, shape=img.shape[:2])
        img[rr, cc] = opposite_color

        # Then draw a smaller polygon on top of the 1st polygon in its middle.
        return self._draw_object(img, obj, is_goal=True)

    def _find_opposite_color(self, color: tuple) -> tuple:
        '''Return the exact opposite of the given color.'''
        return (255 - color[0], 255 - color[1], 255 - color[2])


if __name__ == '__main__':
    import sys

    import machine_common_sense as mcs
    controller = mcs.create_controller(
        unity_cache_version='dev',
        config_file_or_dict={'metadata': 'level1',
                             'video_enabled': True})

    scene_data, _ = mcs.load_scene_json_file(sys.argv[1])
    controller.start_scene(scene_data)
    controller.end_scene()



import json
import logging
import os
import pathlib
from time import perf_counter
from typing import Union

from numpyencoder import NumpyEncoder

from machine_common_sense.step_metadata import StepMetadata

from .controller_events import (AbstractControllerSubscriber, AfterStepPayload,
                                BeforeStepPayload, EndScenePayload,
                                StartScenePayload)
from .goal_metadata import GoalCategory
from .scene_history import SceneHistory
from .stringifier import Stringifier

logger = logging.getLogger(__name__)


class HistoryEventHandler(AbstractControllerSubscriber):

    def __init__(self):
        AbstractControllerSubscriber.__init__(self)
        self.__history_writer = None
        self.__history_item = None

    def on_start_scene(self, payload: StartScenePayload):
        if payload.config.is_history_enabled():

            # Ensure the previous scene history writer has saved its file.
            if self.__history_writer:
                self.__history_writer.check_file_written()

            hist_info = {
                payload.config.CONFIG_EVALUATION_NAME:
                    payload.config.get_evaluation_name(),
                payload.config.CONFIG_METADATA_TIER:
                    payload.config.get_metadata_tier().value,
                payload.config.CONFIG_TEAM:
                    payload.config.get_team()
            }

            # Create a new scene history writer with each new scene (config
            # data) so we always create a new, separate scene history file.
            # Also, ensure previous history item is cleared.
            self.__history_writer = HistoryWriter(payload.scene_config,
                                                  hist_info,
                                                  payload.timestamp)
            init_history = SceneHistory(
                step=payload.step_number,
                action='Initialize',
                args={},
                # We don't have the equivalent of params at this point
                params=None,
                output=payload.step_output.copy_without_depth_or_images(),
                delta_time_millis=0)
            self.__history_writer.add_step(init_history)
            self.__history_item = None

    def on_before_step(self, payload: BeforeStepPayload):
        if payload.config.is_history_enabled():
            if payload.step_number == 0:
                self.__history_writer.init_timer()
            if payload.step_number > 0:
                self.__history_writer.add_step(self.__history_item)
                self.__history_item = None

    def on_after_step(self, payload: AfterStepPayload):
        output: StepMetadata = \
            payload.step_output.copy_without_depth_or_images()

        self.__history_item = SceneHistory(
            step=payload.step_number,
            action=payload.ai2thor_action,
            args=payload.action_kwargs,
            params=payload.step_params,
            output=output,
            delta_time_millis=0,
            target_is_visible_at_start=(
                payload.step_metadata.metadata.get(
                    'targetIsVisibleAtStart')))

    def on_end_scene(self, payload: EndScenePayload):
        if (
            self.__history_writer is not None and
            payload.config.is_history_enabled()
        ):
            self.__history_writer.add_step(self.__history_item)

            # Loop back and fill out previous steps with retrospective report
            if payload.report is not None:
                for step in self.__history_writer.current_steps:
                    current_step = step.get("step")

                    report_step = (
                        payload.report.get(current_step) or
                        payload.report.get(str(current_step))
                    )

                    if (report_step is not None):
                        # Use classification and confidence rather than rating
                        # and score to be compatible with old history files.
                        step["classification"] = report_step.get("rating")
                        step["confidence"] = report_step.get("score")
                        step["violations_xy_list"] = report_step.get(
                            "violations_xy_list")
                        step["internal_state"] = report_step.get(
                            "internal_state")

            self.__history_writer.write_history_file(
                payload.rating, payload.score)

    def _get_filename_without_timestamp(self, filepath: pathlib.Path):
        return filepath.stem[:-16] + filepath.suffix


class HistoryWriter(object):
    HISTORY_DIRECTORY = "SCENE_HISTORY"

    def __init__(self, scene_config_data=None, hist_info=None, timestamp=''):
        if hist_info is None:
            hist_info = {}
        self.info_obj = hist_info
        self.current_steps = []
        self.end_score = {}
        self.scene_history_file = None
        self.history_obj = {}
        self.last_step_time_millis = perf_counter() * 1000

        if not os.path.exists(self.HISTORY_DIRECTORY):
            logger.debug(f"Making history directory {self.HISTORY_DIRECTORY}")
            os.makedirs(self.HISTORY_DIRECTORY)

        scene_name = scene_config_data.name
        prefix_directory = None
        if '/' in scene_name:
            prefix, _ = scene_name.rsplit('/', 1)
            prefix_directory = os.path.join(self.HISTORY_DIRECTORY, prefix)
            if not os.path.exists(prefix_directory):
                logger.debug(f"Making prefix directory {prefix_directory}")
                os.makedirs(prefix_directory)

        if (not scene_config_data.screenshot):
            self.scene_history_file = os.path.join(
                self.HISTORY_DIRECTORY, scene_config_data.name.replace(
                    '.json', '') + "-" + timestamp + ".json")

        self.info_obj['name'] = scene_config_data.name.replace(
            '.json', '')
        self.info_obj['timestamp'] = timestamp

    def write_file(self):
        if self.scene_history_file:
            logger.info(f"Saving history file {self.scene_history_file}")
            with open(self.scene_history_file, "a+") as history_file:
                history = self._round_all(self.history_obj)
                history_file.write(
                    json.dumps(
                        history,
                        cls=NumpyEncoder))

    # This could easily be moved to a more accessible utility file.
    def _round_all(self, obj, num_digits=4):
        if isinstance(obj, float):
            obj = round(obj, num_digits)
        elif isinstance(obj, list):
            obj = [self._round_all(item) for item in obj]
        elif isinstance(obj, dict):
            new_obj = {
                subkey: self._round_all(val) for subkey,
                val in obj.items()}
            obj = new_obj
        elif isinstance(obj, tuple):
            obj = list(obj)
            obj = tuple(self._round_all(obj))
        return obj

    def update_history_output(
            self,
            history: SceneHistory) -> SceneHistory:
        """ filter out images from the step history data and
            object lists and action list, incorporate
            additional target info if needed """
        if history.output:
            history.output.action_list = None
            history.output.structural_object_list = None

            metadata = history.output.goal.metadata or {}
            # Different goal categories may use different property names
            target_names = ['target', 'targets', 'target_1', 'target_2']
            for target_name in target_names:
                # Some properties may be dicts, and some may be lists of dicts
                targets = metadata.get(target_name) or []
                targets = targets if isinstance(targets, list) else [targets]
                for target in targets:
                    # Backwards compatibility: target used to have image data
                    if 'image' in target:
                        del target['image']

                    # Save target position at each step for history file/
                    # scorecard purposes (accounts for tasks where the
                    # target position changes)
                    target_info = next(
                        (obj for obj in history.output.object_list
                         if obj.uuid == target.get('id')), None)

                    if target_info:
                        target["position"] = target_info.position

            history.output.object_list = None
        return history

    def init_timer(self):
        """Initialize the step timer.  Should be called when first command is
            sent to controller"""
        self.last_step_time_millis = perf_counter() * 1000

    def add_step(self, step_obj: SceneHistory):
        """Add a new step to the array of history steps"""
        current_time = perf_counter() * 1000
        if step_obj is not None:
            step_obj.delta_time_millis = (
                current_time - self.last_step_time_millis)
            self.last_step_time_millis = current_time
            if step_obj.output:
                step_obj.target_visible = self.is_target_visible(step_obj)
            logger.debug("Adding history step")
            self.current_steps.append(
                dict(self.update_history_output(step_obj)))

    def is_target_visible(self, history: SceneHistory) -> Union[bool, list]:
        """Determine the current visibility of the target object, if any. In
        scenes with multiple targets, returns true if any target is visible."""
        visible_targets = []
        # Different goal categories may use different property names
        target_names = ['target', 'targets', 'target_1', 'target_2']
        metadata = history.output.goal.metadata or {}
        for target_name in target_names:
            # Some properties may be dicts, and some may be lists of dicts
            targets = metadata.get(target_name) or []
            targets = targets if isinstance(targets, list) else [targets]
            for target in targets:
                for hist_obj in history.output.object_list:
                    if hist_obj.uuid == target.get('id') and hist_obj.visible:
                        visible_targets.append(target.get('id'))
        # Multi-retrieval goals will return a list of all visible target IDs.
        if history.output.goal.category == GoalCategory.MULTI_RETRIEVAL.value:
            return list(sorted(visible_targets))
        # Other goals will return a boolean (backwards compatibility).
        return bool(visible_targets)

    def write_history_file(self, rating, score):
        """ Add the end score obj, create the object
            that will be written to file"""
        # Use classification and confidence rather than rating and score to be
        # compatible with old history files.
        self.end_score["classification"] = rating
        self.end_score["confidence"] = str(score)

        self.history_obj["info"] = self.info_obj
        self.history_obj["steps"] = self.current_steps
        self.history_obj["score"] = self.end_score

        self.write_file()

    def check_file_written(self):
        """ Will check to see if the file has been written, if not,
            it will write out what is currently in the history object"""
        if not os.path.exists(self.scene_history_file):
            self.write_history_file("", "")

    def __str__(self):
        return Stringifier.class_to_str(self)



import unittest
from types import SimpleNamespace

import numpy

import machine_common_sense as mcs
from machine_common_sense.config_manager import (ConfigManager,
                                                 FloorPartitionConfig,
                                                 MetadataTier,
                                                 SceneConfiguration,
                                                 Vector2dInt, Vector3d)
from machine_common_sense.controller_output_handler import (
    ControllerOutputHandler, SceneEvent)
from machine_common_sense.goal_metadata import GoalMetadata
from machine_common_sense.logging_config import LoggingConfig

# ignore printing of errors for unit testing
LoggingConfig.init_logging(LoggingConfig.get_no_logging_config())


class TestControllerOutputHandler(unittest.TestCase):

    def setUp(self):
        self._config = ConfigManager(config_file_or_dict={})
        self._config._config[
            ConfigManager.CONFIG_DEFAULT_SECTION
        ] = {}

    @classmethod
    def tearDownClass(cls) -> None:
        pass

    def create_mock_scene_event(self, mock_scene_event_data):
        # Wrap the dict in a SimpleNamespace object to permit property access
        # with dotted notation since the actual variable is a class, not a
        # dict.
        ''''''
        return SimpleNamespace(**mock_scene_event_data)

    def create_wrap_output_scene_event(self):
        image_data = numpy.array([[0]], dtype=numpy.uint8)
        depth_data = numpy.array([[0.2, 0.4], [0.6, 0.8]], dtype=numpy.float32)
        object_mask_data = numpy.array([[192]], dtype=numpy.uint8)

        return {
            "events": [self.create_mock_scene_event({
                "depth_frame": depth_data,
                "frame": image_data,
                "instance_segmentation_frame": object_mask_data,
                "object_id_to_color": {
                    "testId": (12, 34, 56),
                    "testWallId": (101, 102, 103)
                }
            })],
            "metadata": {
                "agent": {
                    "cameraHorizon": 12.34,
                    "position": {
                        "x": 0.12,
                        "y": -0.23,
                        "z": 4.5
                    },
                    "rotation": {
                        "x": 1.111,
                        "y": 2.222,
                        "z": 3.333
                    }
                },
                "cameraPosition": {
                    "y": 0.1234
                },
                "resolvedObject": 'testResolvedId',
                "resolvedReceptacle": '',
                "clippingPlaneFar": 150,
                "clippingPlaneNear": 0,
                "colors": [{
                    "name": "thing1",
                    "color": [0, 1, 2]
                }, {
                    "name": "thing2",
                    "color": [3, 4, 5]
                }],
                "fov": 42.5,
                "lastActionStatus": "SUCCESSFUL",
                "lastActionSuccess": True,
                "objects": [{
                    "colorsFromMaterials": ["c1"],
                    "direction": {
                        "x": 90,
                        "y": -30,
                        "z": 0
                    },
                    "distance": 1.5,
                    "distanceXZ": 1.1,
                    "isPickedUp": False,
                    "mass": 12.34,
                    "objectBounds": {
                        "objectBoundsCorners": [
                            "p1",
                            "p2",
                            "p3",
                            "p4",
                            "p5",
                            "p6",
                            "p7",
                            "p8",
                        ]
                    },
                    "objectId": "testId",
                    "position": {
                        "x": 10,
                        "y": 11,
                        "z": 12
                    },
                    "rotation": {
                        "x": 1.0,
                        "y": 2.0,
                        "z": 3.0
                    },
                    "salientMaterials": ["Wood"],
                    "shape": "shape",
                    "visibleInCamera": True,
                    "isOpen": False,
                    "openable": False,
                    "locked": False,
                    "associatedWithAgent": "",
                    "simulationAgentHeldObject": "",
                    "simulationAgentIsHoldingHeldObject": False
                }, {
                    "colorsFromMaterials": [],
                    "direction": {
                        "x": -90,
                        "y": 180,
                        "z": 270
                    },
                    "distance": 2.5,
                    "distanceXZ": 2.0,
                    "isPickedUp": False,
                    "mass": 34.56,
                    "objectBounds": {
                        "objectBoundsCorners": [
                            "pA",
                            "pB",
                            "pC",
                            "pD",
                            "pE",
                            "pF",
                            "pG",
                            "pH",
                        ]
                    },
                    "objectId": "testIdHidden",
                    "position": {
                        "x": -3,
                        "y": -2,
                        "z": -1
                    },
                    "rotation": {
                        "x": 11.0,
                        "y": 12.0,
                        "z": 13.0
                    },
                    "salientMaterials": ["Wood"],
                    "shape": "shapeHidden",
                    "visibleInCamera": False,
                    "isOpen": False,
                    "openable": False,
                    "locked": False,
                    "associatedWithAgent": "",
                    "simulationAgentHeldObject": "",
                    "simulationAgentIsHoldingHeldObject": False
                }],
                "structuralObjects": [{
                    "colorsFromMaterials": ["c2"],
                    "direction": {
                        "x": 180,
                        "y": -60,
                        "z": 0
                    },
                    "distance": 2.5,
                    "distanceXZ": 2.2,
                    "isPickedUp": False,
                    "mass": 56.78,
                    "objectBounds": {
                        "objectBoundsCorners": [
                            "p11",
                            "p12",
                            "p13",
                            "p14",
                            "p15",
                            "p16",
                            "p17",
                            "p18",
                        ]
                    },
                    "objectId": "testWallId",
                    "position": {
                        "x": 20,
                        "y": 21,
                        "z": 22
                    },
                    "rotation": {
                        "x": 4.0,
                        "y": 5.0,
                        "z": 6.0
                    },
                    "salientMaterials": ["Ceramic"],
                    "shape": "structure",
                    "visibleInCamera": True,
                    "isOpen": False,
                    "openable": False,
                    "locked": False,
                    "associatedWithAgent": "",
                    "simulationAgentHeldObject": "",
                    "simulationAgentIsHoldingHeldObject": False
                }, {
                    "colorsFromMaterials": [],
                    "direction": {
                        "x": -180,
                        "y": 60,
                        "z": 90
                    },
                    "distance": 3.5,
                    "distanceXZ": 3.3,
                    "isPickedUp": False,
                    "mass": 78.90,
                    "objectBounds": {
                        "objectBoundsCorners": [
                            "pAA",
                            "pBB",
                            "pCC",
                            "pDD",
                            "pEE",
                            "pFF",
                            "pGG",
                            "pHH",
                        ]
                    },
                    "objectId": "testWallIdHidden",
                    "position": {
                        "x": 30,
                        "y": 31,
                        "z": 32
                    },
                    "rotation": {
                        "x": 14.0,
                        "y": 15.0,
                        "z": 16.0
                    },
                    "salientMaterials": ["Ceramic"],
                    "shape": "structureHidden",
                    "visibleInCamera": False,
                    "isOpen": False,
                    "openable": False,
                    "locked": False,
                    "associatedWithAgent": "",
                    "simulationAgentHeldObject": "",
                    "simulationAgentIsHoldingHeldObject": False
                }]
            }
        }, image_data, depth_data, object_mask_data

    def create_retrieve_object_list_scene_event(self):
        return {
            "events": [self.create_mock_scene_event({
                "object_id_to_color": {
                    "testId1": (12, 34, 56),
                    "testId2": (98, 76, 54),
                    "testId3": (101, 102, 103)
                }
            })],
            "metadata": {
                "objects": [{
                    "colorsFromMaterials": ["c1"],
                    "direction": {
                        "x": 0,
                        "y": 0,
                        "z": 0
                    },
                    "distance": 0,
                    "distanceXZ": 0,
                    "isPickedUp": True,
                    "mass": 1,
                    "objectId": "testId1",
                    "position": {
                        "x": 1,
                        "y": 1,
                        "z": 2
                    },
                    "rotation": {
                        "x": 1.0,
                        "y": 2.0,
                        "z": 3.0
                    },
                    "salientMaterials": [],
                    "shape": "shape1",
                    "visibleInCamera": True,
                    "isOpen": False,
                    "openable": False,
                    "locked": False,
                    "associatedWithAgent": "",
                    "simulationAgentHeldObject": "",
                    "simulationAgentIsHoldingHeldObject": False
                }, {
                    "colorsFromMaterials": ["c2", "c3"],
                    "direction": {
                        "x": 90,
                        "y": -30,
                        "z": 0
                    },
                    "distance": 1.5,
                    "distanceXZ": 1.1,
                    "isPickedUp": False,
                    "mass": 12.34,
                    "objectBounds": {
                        "objectBoundsCorners": [
                            "p1",
                            "p2",
                            "p3",
                            "p4",
                            "p5",
                            "p6",
                            "p7",
                            "p8",
                        ]
                    },
                    "objectId": "testId2",
                    "position": {
                        "x": 1,
                        "y": 2,
                        "z": 3
                    },
                    "rotation": {
                        "x": 1.0,
                        "y": 2.0,
                        "z": 3.0
                    },
                    "salientMaterials": ["Foobar", "Metal", "Plastic"],
                    "shape": "shape2",
                    "visibleInCamera": True,
                    "isOpen": False,
                    "openable": False,
                    "locked": False,
                    "associatedWithAgent": "agent_test",
                    "simulationAgentHeldObject": "",
                    "simulationAgentIsHoldingHeldObject": False
                }, {
                    "colorsFromMaterials": [],
                    "direction": {
                        "x": -90,
                        "y": 180,
                        "z": 270
                    },
                    "distance": 2.5,
                    "distanceXZ": 2,
                    "isPickedUp": False,
                    "mass": 34.56,
                    "objectBounds": {
                        "objectBoundsCorners": [
                            "pA",
                            "pB",
                            "pC",
                            "pD",
                            "pE",
                            "pF",
                            "pG",
                            "pH",
                        ]
                    },
                    "objectId": "testId3",
                    "position": {
                        "x": -3,
                        "y": -2,
                        "z": -1
                    },
                    "rotation": {
                        "x": 11.0,
                        "y": 12.0,
                        "z": 13.0
                    },
                    "salientMaterials": ["Wood"],
                    "shape": "shape3",
                    "visibleInCamera": False,
                    "isOpen": False,
                    "openable": False,
                    "locked": False,
                    "associatedWithAgent": "",
                    "simulationAgentHeldObject": "held_test",
                    "simulationAgentIsHoldingHeldObject": True
                }]
            }
        }

    def test_retrieve_return_status(self):
        mock_scene_event_data = {
            "metadata": {
                "lastActionStatus": "SUCCESSFUL"
            },
            "events": []
        }

        mock_event = self.create_mock_scene_event(mock_scene_event_data)
        scene_event = SceneEvent(
            self._config, {}, mock_event, 0)

        actual = scene_event.return_status
        self.assertEqual(actual, mcs.ReturnStatus.SUCCESSFUL.name)

        mock_scene_event_data = {
            "metadata": {
                "lastActionStatus": "FAILED"
            },
            "events": []
        }

        mock_event = self.create_mock_scene_event(mock_scene_event_data)
        scene_event = SceneEvent(
            self._config, {}, mock_event, 0)
        actual = scene_event.return_status
        self.assertEqual(actual, mcs.ReturnStatus.FAILED.name)

        mock_scene_event_data = {
            "metadata": {
                "lastActionStatus": "INVALID_STATUS"
            },
            "events": []
        }

        mock_event = self.create_mock_scene_event(mock_scene_event_data)
        scene_event = SceneEvent(
            self._config, {}, mock_event, 0)
        actual = scene_event.return_status
        self.assertEqual(actual, mcs.ReturnStatus.UNDEFINED.name)

        mock_scene_event_data = {
            "metadata": {
                "lastActionStatus": None
            },
            "events": []
        }

        mock_event = self.create_mock_scene_event(mock_scene_event_data)
        scene_event = SceneEvent(
            self._config, {}, mock_event, 0)
        actual = scene_event.return_status
        self.assertEqual(actual, mcs.ReturnStatus.UNDEFINED.name)

    def test_retrieve_head_tilt(self):
        mock_scene_event_data = {
            "metadata": {
                "agent": {
                    "cameraHorizon": 12.34
                }
            },
            "events": []
        }

        mock_event = self.create_mock_scene_event(mock_scene_event_data)

        scene_event = SceneEvent(
            self._config, {}, mock_event, 0)

        actual = scene_event.head_tilt
        self.assertEqual(actual, 12.34)

        mock_scene_event_data = {
            "metadata": {
                "agent": {
                    "cameraHorizon": -56.78
                }
            },
            "events": []
        }

        mock_event = self.create_mock_scene_event(mock_scene_event_data)
        scene_event = SceneEvent(
            self._config, {}, mock_event, 0)
        actual = scene_event.head_tilt
        self.assertEqual(actual, -56.78)

    def test_wrap_output(self):
        self._config.set_metadata_tier(
            MetadataTier.DEFAULT.value)
        (
            mock_scene_event_data,
            image_data,
            depth_data,
            object_mask_data
        ) = self.create_wrap_output_scene_event()
        mock_event = self.create_mock_scene_event(mock_scene_event_data)

        coh = ControllerOutputHandler(self._config)
        coh.set_scene_config(SceneConfiguration(holes=[
            Vector2dInt(x=0, z=0), Vector2dInt(x=1, z=2), Vector2dInt(x=9, z=8)
        ], lava=[
            Vector2dInt(x=3, z=3), Vector2dInt(x=7, z=5), Vector2dInt(x=4, z=6)
        ]))
        goal = GoalMetadata(metadata={
            'target': {'id': 'target_id_1'},
            'targets': [{'id': 'target_id_2'}, {'id': 'target_id_3'}]
        }, triggered_by_target_sequence=['chest_1', 'chest_3'])
        (res, actual) = coh.handle_output(mock_event, goal, 0, 1)

        self.assertEqual(actual.action_list, GoalMetadata.DEFAULT_ACTIONS)
        self.assertEqual(actual.camera_aspect_ratio, (600, 400))
        self.assertEqual(actual.camera_clipping_planes, (0, 150))
        self.assertEqual(actual.camera_field_of_view, 42.5)
        self.assertEqual(actual.camera_height, 0.1234)
        self.assertEqual(actual.goal.metadata, goal.metadata)
        self.assertEqual(actual.goal.triggered_by_target_sequence,
                         goal.triggered_by_target_sequence)
        self.assertEqual(actual.habituation_trial, None)
        self.assertEqual(actual.head_tilt, 12.34)
        self.assertEqual(actual.holes, [(0, 0), (1, 2), (9, 8)])
        self.assertEqual(actual.lava, [
            (2.5, 2.5, 3.5, 3.5), (6.5, 4.5, 7.5, 5.5), (3.5, 5.5, 4.5, 6.5)
        ])
        self.assertEqual(actual.position, {'x': 0.12, 'y': -0.23, 'z': 4.5})
        self.assertEqual(actual.resolved_object, 'testResolvedId')
        self.assertEqual(actual.resolved_receptacle, '')
        self.assertEqual(actual.rotation, 2.222)
        self.assertEqual(
            actual.return_status,
            mcs.ReturnStatus.SUCCESSFUL.value)
        self.assertEqual(actual.segmentation_colors, [
            {'objectId': 'thing1', 'r': 0, 'g': 1, 'b': 2},
            {'objectId': 'thing2', 'r': 3, 'g': 4, 'b': 5}
        ])
        self.assertEqual(actual.step_number, 0)

        self.assertEqual(len(actual.object_list), 1)
        self.assertEqual(actual.object_list[0].uuid, "testId")
        self.assertEqual(actual.object_list[0].segment_color, {
            "r": 12,
            "g": 34,
            "b": 56
        })
        self.assertEqual(
            actual.object_list[0].dimensions, [
                "p1", "p2", "p3", "p4", "p5", "p6", "p7", "p8"])
        self.assertEqual(actual.object_list[0].direction, {
            "x": 90,
            "y": -30,
            "z": 0
        })
        self.assertEqual(actual.object_list[0].distance, 11.0)
        self.assertEqual(actual.object_list[0].distance_in_steps, 11.0)
        self.assertEqual(actual.object_list[0].distance_in_world, 1.5)
        self.assertEqual(actual.object_list[0].held, False)
        self.assertEqual(actual.object_list[0].mass, 12.34)
        self.assertEqual(actual.object_list[0].material_list, ["WOOD"])
        self.assertEqual(
            actual.object_list[0].position, {
                "x": 10, "y": 11, "z": 12})
        self.assertEqual(
            actual.object_list[0].rotation, {
                "x": 1, "y": 2, "z": 3})
        self.assertEqual(actual.object_list[0].shape, 'shape')
        self.assertEqual(actual.object_list[0].state_list, [])
        self.assertEqual(actual.object_list[0].texture_color_list, ['c1'])
        self.assertEqual(actual.object_list[0].visible, True)
        self.assertEqual(actual.object_list[0].associated_with_agent, "")
        self.assertEqual(
            actual.object_list[0].simulation_agent_held_object, "")
        self.assertEqual(
            actual.object_list[0].simulation_agent_is_holding_held_object,
            False)

        self.assertEqual(len(actual.structural_object_list), 1)
        self.assertEqual(actual.structural_object_list[0].uuid, "testWallId")
        self.assertEqual(actual.structural_object_list[0].segment_color, {
            "r": 101,
            "g": 102,
            "b": 103
        })
        self.assertEqual(
            actual.structural_object_list[0].dimensions,
            ["p11", "p12", "p13", "p14", "p15", "p16", "p17", "p18"]
        )
        self.assertEqual(actual.structural_object_list[0].direction, {
            "x": 180,
            "y": -60,
            "z": 0
        })
        self.assertEqual(actual.structural_object_list[0].distance, 22.0)
        self.assertEqual(
            actual.structural_object_list[0].distance_in_steps, 22.0)
        self.assertEqual(
            actual.structural_object_list[0].distance_in_world, 2.5)
        self.assertEqual(actual.structural_object_list[0].held, False)
        self.assertEqual(actual.structural_object_list[0].mass, 56.78)
        self.assertEqual(
            actual.structural_object_list[0].material_list,
            ["CERAMIC"])
        self.assertEqual(
            actual.structural_object_list[0].position, {
                "x": 20, "y": 21, "z": 22})
        self.assertEqual(
            actual.structural_object_list[0].rotation, {
                "x": 4, "y": 5, "z": 6})
        self.assertEqual(actual.structural_object_list[0].shape, 'structure')
        self.assertEqual(actual.structural_object_list[0].state_list, [])
        self.assertEqual(
            actual.structural_object_list[0].texture_color_list,
            ['c2'])
        self.assertEqual(actual.structural_object_list[0].visible, True)
        self.assertEqual(actual.object_list[0].associated_with_agent, "")
        self.assertEqual(
            actual.object_list[0].simulation_agent_held_object, "")
        self.assertEqual(
            actual.object_list[0].simulation_agent_is_holding_held_object,
            False)

        # IF we are at default level, shouldn't depth maps, object masks be
        # restricted?
        self.assertEqual(len(actual.depth_map_list), 0)
        self.assertEqual(len(actual.image_list), 1)
        self.assertEqual(len(actual.object_mask_list), 0)
        '''numpy.testing.assert_almost_equal(
            numpy.array(actual.depth_map_list[0]),
            numpy.array([[30, 60], [90, 120]], dtype=numpy.float32),
            3
        )'''
        self.assertEqual(numpy.array(actual.image_list[0]), image_data)
        '''self.assertEqual(
            numpy.array(
                actual.object_mask_list[0]),
            object_mask_data)
        '''

    def test_wrap_output_with_config_metadata_level2(self):
        self._config.set_metadata_tier('level2')
        (
            mock_scene_event_data,
            image_data,
            depth_data,
            object_mask_data
        ) = self.create_wrap_output_scene_event()
        mock_event = self.create_mock_scene_event(mock_scene_event_data)

        coh = ControllerOutputHandler(self._config)
        coh.set_scene_config(SceneConfiguration(holes=[
            Vector2dInt(x=0, z=0), Vector2dInt(x=1, z=2), Vector2dInt(x=9, z=8)
        ], lava=[
            Vector2dInt(x=3, z=3), Vector2dInt(x=7, z=5), Vector2dInt(x=4, z=6)
        ]))
        goal = GoalMetadata(metadata={
            'target': {'id': 'target_id_1'},
            'targets': [{'id': 'target_id_2'}, {'id': 'target_id_3'}]
        }, triggered_by_target_sequence=['chest_1', 'chest_3'])
        (res, actual) = coh.handle_output(mock_event, goal, 0, 1)

        self.assertEqual(actual.action_list, GoalMetadata.DEFAULT_ACTIONS)
        self.assertEqual(actual.camera_aspect_ratio, (600, 400))
        self.assertEqual(actual.camera_clipping_planes, (0, 150))
        self.assertEqual(actual.camera_field_of_view, 42.5)
        self.assertEqual(actual.camera_height, 0.1234)
        self.assertEqual(actual.goal.metadata, {
            'target': {},
            'targets': [],
        })
        self.assertEqual(actual.goal.triggered_by_target_sequence, None)
        self.assertEqual(actual.habituation_trial, None)
        self.assertEqual(actual.head_tilt, 12.34)
        self.assertEqual(actual.position, None)
        self.assertEqual(actual.resolved_object, None)
        self.assertEqual(actual.resolved_receptacle, None)
        self.assertEqual(actual.rotation, None)
        self.assertEqual(
            actual.return_status,
            mcs.ReturnStatus.SUCCESSFUL.value)
        self.assertEqual(actual.segmentation_colors, [])
        self.assertEqual(actual.step_number, 0)
        self.assertEqual(actual.holes, None)
        self.assertEqual(actual.lava, None)
        # Correct object metadata properties tested elsewhere
        self.assertEqual(len(actual.object_list), 0)
        self.assertEqual(len(actual.structural_object_list), 0)

        # depth map list used to be 0, but should be 1 based on config level2
        self.assertEqual(len(actual.depth_map_list), 1)
        self.assertEqual(len(actual.image_list), 1)
        # object mask list used to be 0, but should be 1 based on config level2
        self.assertEqual(len(actual.object_mask_list), 1)
        # self.assertEqual(
        #     numpy.array(
        #         actual.depth_map_list[0]),
        #     depth_data)
        self.assertEqual(numpy.array(actual.image_list[0]), image_data)
        # self.assertEqual(
        #     numpy.array(
        #         actual.depth_map_list[0]),
        #     object_mask_data)

    def test_wrap_output_with_config_metadata_level1(self):
        self._config.set_metadata_tier(
            MetadataTier.LEVEL_1.value)
        (
            mock_scene_event_data,
            image_data,
            depth_data,
            object_mask_data
        ) = self.create_wrap_output_scene_event()

        mock_event = self.create_mock_scene_event(mock_scene_event_data)

        coh = ControllerOutputHandler(self._config)
        coh.set_scene_config(SceneConfiguration(holes=[
            Vector2dInt(x=0, z=0), Vector2dInt(x=1, z=2), Vector2dInt(x=9, z=8)
        ], lava=[
            Vector2dInt(x=3, z=3), Vector2dInt(x=7, z=5), Vector2dInt(x=4, z=6)
        ]))
        goal = GoalMetadata(metadata={
            'target': {'id': 'target_id_1'},
            'targets': [{'id': 'target_id_2'}, {'id': 'target_id_3'}]
        }, triggered_by_target_sequence=['chest_1', 'chest_3'])
        (res, actual) = coh.handle_output(mock_event, goal, 0, 1)

        self.assertEqual(actual.action_list, GoalMetadata.DEFAULT_ACTIONS)
        self.assertEqual(actual.camera_aspect_ratio, (600, 400))
        self.assertEqual(actual.camera_clipping_planes, (0, 150))
        self.assertEqual(actual.camera_field_of_view, 42.5)
        self.assertEqual(actual.camera_height, 0.1234)
        self.assertEqual(actual.goal.metadata, {
            'target': {},
            'targets': [],
        })
        self.assertEqual(actual.goal.triggered_by_target_sequence, None)
        self.assertEqual(actual.habituation_trial, None)
        self.assertEqual(actual.head_tilt, 12.34)
        self.assertEqual(actual.position, None)
        self.assertEqual(actual.resolved_object, None)
        self.assertEqual(actual.resolved_receptacle, None)
        self.assertEqual(actual.rotation, None)
        self.assertEqual(
            actual.return_status,
            mcs.ReturnStatus.SUCCESSFUL.value)
        self.assertEqual(actual.segmentation_colors, [])
        self.assertEqual(actual.step_number, 0)
        self.assertEqual(actual.holes, None)
        self.assertEqual(actual.lava, None)
        # Correct object metadata properties tested elsewhere
        self.assertEqual(len(actual.object_list), 0)
        self.assertEqual(len(actual.structural_object_list), 0)

        self.assertEqual(len(actual.depth_map_list), 1)
        self.assertEqual(len(actual.image_list), 1)
        self.assertEqual(len(actual.object_mask_list), 0)

    def test_wrap_output_with_config_metadata_none(self):
        self._config.set_metadata_tier('none')
        (
            mock_scene_event_data,
            image_data,
            depth_data,
            object_mask_data
        ) = self.create_wrap_output_scene_event()

        mock_event = self.create_mock_scene_event(mock_scene_event_data)

        coh = ControllerOutputHandler(self._config)
        coh.set_scene_config(SceneConfiguration(holes=[
            Vector2dInt(x=0, z=0), Vector2dInt(x=1, z=2), Vector2dInt(x=9, z=8)
        ], lava=[
            Vector2dInt(x=3, z=3), Vector2dInt(x=7, z=5), Vector2dInt(x=4, z=6)
        ]))
        goal = GoalMetadata(metadata={
            'target': {'id': 'target_id_1'},
            'targets': [{'id': 'target_id_2'}, {'id': 'target_id_3'}]
        }, triggered_by_target_sequence=['chest_1', 'chest_3'])
        (res, actual) = coh.handle_output(mock_event, goal, 0, 1)

        self.assertEqual(actual.action_list, GoalMetadata.DEFAULT_ACTIONS)
        self.assertEqual(actual.camera_aspect_ratio, (600, 400))
        self.assertEqual(actual.camera_clipping_planes, (0, 150))
        self.assertEqual(actual.camera_field_of_view, 42.5)
        self.assertEqual(actual.camera_height, 0.1234)
        self.assertEqual(actual.goal.metadata, {
            'target': {},
            'targets': [],
        })
        self.assertEqual(actual.triggered_by_sequence_incorrect, None)
        self.assertEqual(actual.habituation_trial, None)
        self.assertEqual(actual.head_tilt, 12.34)
        self.assertEqual(actual.position, None)
        self.assertEqual(actual.resolved_object, None)
        self.assertEqual(actual.resolved_receptacle, None)
        self.assertEqual(actual.rotation, None)
        self.assertEqual(
            actual.return_status,
            mcs.ReturnStatus.SUCCESSFUL.value)
        self.assertEqual(actual.segmentation_colors, [])
        self.assertEqual(actual.step_number, 0)

        # Correct object metadata properties tested elsewhere
        self.assertEqual(len(actual.object_list), 0)
        self.assertEqual(len(actual.structural_object_list), 0)

        self.assertEqual(len(actual.depth_map_list), 0)
        self.assertEqual(len(actual.image_list), 1)
        self.assertEqual(len(actual.object_mask_list), 0)

    def test_wrap_output_with_config_metadata_oracle_non_restrict(self):
        self._config.set_metadata_tier(
            MetadataTier.ORACLE.value)
        (
            mock_scene_event_data,
            image_data,
            depth_data,
            object_mask_data
        ) = self.create_wrap_output_scene_event()

        mock_event = self.create_mock_scene_event(mock_scene_event_data)

        coh = ControllerOutputHandler(self._config)
        coh.set_scene_config(SceneConfiguration(holes=[
            Vector2dInt(x=0, z=0), Vector2dInt(x=1, z=2), Vector2dInt(x=9, z=8)
        ], lava=[
            Vector2dInt(x=3, z=3), Vector2dInt(x=7, z=5), Vector2dInt(x=4, z=6)
        ]))
        goal = GoalMetadata(metadata={
            'target': {'id': 'target_id_1'},
            'targets': [{'id': 'target_id_2'}, {'id': 'target_id_3'}]
        }, triggered_by_target_sequence=['chest_1', 'chest_3'])
        (actual, res) = coh.handle_output(mock_event, goal, 0, 1)

        self.assertEqual(actual.action_list, GoalMetadata.DEFAULT_ACTIONS)
        self.assertEqual(actual.camera_aspect_ratio, (600, 400))
        self.assertEqual(actual.camera_clipping_planes, (0, 150))
        self.assertEqual(actual.camera_field_of_view, 42.5)
        self.assertEqual(actual.camera_height, 0.1234)
        self.assertEqual(actual.goal.metadata, goal.metadata)
        self.assertEqual(actual.goal.triggered_by_target_sequence,
                         goal.triggered_by_target_sequence)
        self.assertEqual(actual.habituation_trial, None)
        self.assertEqual(actual.head_tilt, 12.34)
        self.assertEqual(actual.position, {'x': 0.12, 'y': -0.23, 'z': 4.5})
        self.assertEqual(actual.resolved_object, 'testResolvedId')
        self.assertEqual(actual.resolved_receptacle, '')
        self.assertEqual(actual.rotation, 2.222)
        self.assertEqual(
            actual.return_status,
            mcs.ReturnStatus.SUCCESSFUL.value)
        self.assertEqual(actual.segmentation_colors, [
            {'objectId': 'thing1', 'r': 0, 'g': 1, 'b': 2},
            {'objectId': 'thing2', 'r': 3, 'g': 4, 'b': 5}
        ])
        self.assertEqual(actual.step_number, 0)
        self.assertEqual(actual.holes, [(0, 0), (1, 2), (9, 8)])
        self.assertEqual(actual.lava, [
            (2.5, 2.5, 3.5, 3.5), (6.5, 4.5, 7.5, 5.5), (3.5, 5.5, 4.5, 6.5)
        ])
        # Correct object metadata properties tested elsewhere
        self.assertEqual(len(actual.object_list), 2)
        self.assertEqual(len(actual.structural_object_list), 2)

        self.assertEqual(len(actual.depth_map_list), 1)
        self.assertEqual(len(actual.image_list), 1)
        self.assertEqual(len(actual.object_mask_list), 1)
        numpy.testing.assert_almost_equal(
            numpy.array(actual.depth_map_list[0]),
            numpy.array([[30, 60], [90, 120]], dtype=numpy.float32),
            3
        )
        self.assertEqual(numpy.array(actual.image_list[0]), image_data)
        self.assertEqual(
            numpy.array(
                actual.object_mask_list[0]),
            object_mask_data)

    def test_wrap_output_with_config_metadata_oracle(self):
        self._config.set_metadata_tier(
            MetadataTier.ORACLE.value)
        (
            mock_scene_event_data,
            image_data,
            depth_data,
            object_mask_data
        ) = self.create_wrap_output_scene_event()

        mock_event = self.create_mock_scene_event(mock_scene_event_data)

        coh = ControllerOutputHandler(self._config)
        coh.set_scene_config(SceneConfiguration(holes=[
            Vector2dInt(x=0, z=0), Vector2dInt(x=1, z=2), Vector2dInt(x=9, z=8)
        ], lava=[
            Vector2dInt(x=3, z=3), Vector2dInt(x=7, z=5), Vector2dInt(x=4, z=6)
        ]))
        goal = GoalMetadata(metadata={
            'target': {'id': 'target_id_1'},
            'targets': [{'id': 'target_id_2'}, {'id': 'target_id_3'}]
        }, triggered_by_target_sequence=['chest_1', 'chest_3'])
        (res, actual) = coh.handle_output(mock_event, goal, 0, 1)

        self.assertEqual(actual.action_list, GoalMetadata.DEFAULT_ACTIONS)
        self.assertEqual(actual.camera_aspect_ratio, (600, 400))
        self.assertEqual(actual.camera_clipping_planes, (0, 150))
        self.assertEqual(actual.camera_field_of_view, 42.5)
        self.assertEqual(actual.camera_height, 0.1234)
        self.assertEqual(actual.goal.metadata, goal.metadata)
        self.assertEqual(actual.goal.triggered_by_target_sequence,
                         goal.triggered_by_target_sequence)
        self.assertEqual(actual.habituation_trial, None)
        self.assertEqual(actual.head_tilt, 12.34)
        self.assertEqual(actual.position, {'x': 0.12, 'y': -0.23, 'z': 4.5})
        self.assertEqual(actual.resolved_object, 'testResolvedId')
        self.assertEqual(actual.resolved_receptacle, '')
        self.assertEqual(actual.rotation, 2.222)
        self.assertEqual(
            actual.return_status,
            mcs.ReturnStatus.SUCCESSFUL.value)
        self.assertEqual(actual.segmentation_colors, [
            {'objectId': 'thing1', 'r': 0, 'g': 1, 'b': 2},
            {'objectId': 'thing2', 'r': 3, 'g': 4, 'b': 5}
        ])
        self.assertEqual(actual.step_number, 0)
        self.assertEqual(actual.holes, [(0, 0), (1, 2), (9, 8)])
        self.assertEqual(actual.lava, [
            (2.5, 2.5, 3.5, 3.5), (6.5, 4.5, 7.5, 5.5), (3.5, 5.5, 4.5, 6.5)
        ])
        # Correct object metadata properties tested elsewhere
        self.assertEqual(len(actual.object_list), 2)
        self.assertEqual(len(actual.structural_object_list), 2)

        self.assertEqual(len(actual.depth_map_list), 1)
        self.assertEqual(len(actual.image_list), 1)
        self.assertEqual(len(actual.object_mask_list), 1)
        numpy.testing.assert_almost_equal(
            numpy.array(actual.depth_map_list[0]),
            numpy.array([[30, 60], [90, 120]], dtype=numpy.float32),
            3
        )
        self.assertEqual(numpy.array(actual.image_list[0]), image_data)
        self.assertEqual(
            numpy.array(
                actual.object_mask_list[0]),
            object_mask_data)

    def test_wrap_output_with_partition_floor(self):
        self._config.set_metadata_tier(MetadataTier.ORACLE.value)
        (
            mock_scene_event_data,
            image_data,
            depth_data,
            object_mask_data
        ) = self.create_wrap_output_scene_event()
        mock_event = self.create_mock_scene_event(mock_scene_event_data)

        coh = ControllerOutputHandler(self._config)
        coh.set_scene_config(SceneConfiguration(
            partition_floor=FloorPartitionConfig(
                left_half=0.2,
                right_half=0.8
            ),
            room_dimensions=Vector3d(x=30, y=3, z=20)
        ))
        (res, actual) = coh.handle_output(mock_event, GoalMetadata(), 0, 1)

        self.assertEqual(actual.lava, [(-15, 10, -12, -10), (3, 10, 15, -10)])

    def test_save_images(self):
        self._config.set_metadata_tier(
            MetadataTier.ORACLE.value)
        image_data = numpy.array([[0]], dtype=numpy.uint8)
        depth_data = numpy.array([[0.2, 0.4], [0.6, 0.8]], dtype=numpy.float32)
        object_mask_data = numpy.array([[192]], dtype=numpy.uint8)

        mock_scene_event_data = {
            "events": [self.create_mock_scene_event({
                "depth_frame": depth_data,
                "frame": image_data,
                "instance_segmentation_frame": object_mask_data
            })],
            "metadata": {"clippingPlaneFar": 150, "objects": []}
        }
        mock_event = self.create_mock_scene_event(mock_scene_event_data)
        scene_event = SceneEvent(
            self._config, {}, mock_event, 0)

        image_list = scene_event.image_list
        depth_map_list = scene_event.depth_map_list
        object_mask_list = scene_event.object_mask_list

        self.assertEqual(len(image_list), 1)
        self.assertEqual(len(depth_map_list), 1)
        self.assertEqual(len(object_mask_list), 1)

        self.assertEqual(numpy.array(image_list[0]), image_data)
        numpy.testing.assert_almost_equal(
            numpy.array(depth_map_list[0]),
            numpy.array([[30, 60], [90, 120]], dtype=numpy.float32),
            3
        )
        self.assertEqual(numpy.array(object_mask_list[0]), object_mask_data)

    def test_save_images_with_multiple_images(self):
        self._config.set_metadata_tier(
            MetadataTier.ORACLE.value)
        image_data_1 = numpy.array([[64]], dtype=numpy.uint8)
        depth_data_1 = numpy.array(
            [[0.2, 0.4], [0.6, 0.8]],
            dtype=numpy.float32
        )
        object_mask_data_1 = numpy.array([[192]], dtype=numpy.uint8)

        image_data_2 = numpy.array([[32]], dtype=numpy.uint8)
        depth_data_2 = numpy.array(
            [[0.0001, 0.99], [0.25, 0.75]],
            dtype=numpy.float32
        )
        object_mask_data_2 = numpy.array([[160]], dtype=numpy.uint8)

        mock_scene_event_data = {
            "events": [self.create_mock_scene_event({
                "depth_frame": depth_data_1,
                "frame": image_data_1,
                "instance_segmentation_frame": object_mask_data_1
            }), self.create_mock_scene_event({
                "depth_frame": depth_data_2,
                "frame": image_data_2,
                "instance_segmentation_frame": object_mask_data_2
            })],
            "metadata": {"clippingPlaneFar": 150, "objects": []}
        }

        mock_event = self.create_mock_scene_event(mock_scene_event_data)

        scene_event = SceneEvent(
            self._config, {}, mock_event, 0)

        image_list = scene_event.image_list
        depth_map_list = scene_event.depth_map_list
        object_mask_list = scene_event.object_mask_list

        self.assertEqual(len(image_list), 2)
        self.assertEqual(len(depth_map_list), 2)
        self.assertEqual(len(object_mask_list), 2)

        self.assertEqual(numpy.array(image_list[0]), image_data_1)
        numpy.testing.assert_almost_equal(
            numpy.array(depth_map_list[0]),
            numpy.array([[30, 60], [90, 120]], dtype=numpy.float32),
            3
        )
        self.assertEqual(numpy.array(object_mask_list[0]), object_mask_data_1)

        self.assertEqual(numpy.array(image_list[1]), image_data_2)
        numpy.testing.assert_almost_equal(
            numpy.array(depth_map_list[1]),
            numpy.array([[0.015, 148.5], [37.5, 112.5]], dtype=numpy.float32),
            3
        )
        self.assertEqual(numpy.array(object_mask_list[1]), object_mask_data_2)

    def test_retrieve_object_list(self):
        mock_scene_event_data = self.create_retrieve_object_list_scene_event()

        mock_event = self.create_mock_scene_event(mock_scene_event_data)
        scene_event = SceneEvent(
            self._config, SceneConfiguration(), mock_event, 0)
        actual = scene_event.object_list
        self.assertEqual(len(actual), 2)

        self.assertEqual(actual[0].uuid, "testId1")
        self.assertEqual(actual[0].segment_color, {
            "r": 12,
            "g": 34,
            "b": 56
        })
        self.assertEqual(actual[0].dimensions, [])
        self.assertEqual(actual[0].direction, {
            "x": 0,
            "y": 0,
            "z": 0
        })
        self.assertEqual(actual[0].distance, 0)
        self.assertEqual(actual[0].distance_in_steps, 0)
        self.assertEqual(actual[0].distance_in_world, 0)
        self.assertEqual(actual[0].held, True)
        self.assertEqual(actual[0].mass, 1)
        self.assertEqual(actual[0].material_list, [])
        self.assertEqual(actual[0].position, {"x": 1, "y": 1, "z": 2})
        self.assertEqual(actual[0].rotation, {"x": 1, "y": 2, "z": 3})
        self.assertEqual(actual[0].shape, 'shape1')
        self.assertEqual(actual[0].state_list, [])
        self.assertEqual(actual[0].texture_color_list, ['c1'])
        self.assertEqual(actual[0].visible, True)
        self.assertEqual(actual[0].associated_with_agent, "")
        self.assertEqual(
            actual[0].simulation_agent_held_object, "")
        self.assertEqual(
            actual[0].simulation_agent_is_holding_held_object,
            False)

        self.assertEqual(actual[1].uuid, "testId2")
        self.assertEqual(actual[1].segment_color, {
            "r": 98,
            "g": 76,
            "b": 54
        })
        self.assertEqual(
            actual[1].dimensions, [
                "p1", "p2", "p3", "p4", "p5", "p6", "p7", "p8"])
        self.assertEqual(actual[1].direction, {
            "x": 90,
            "y": -30,
            "z": 0
        })
        self.assertEqual(actual[1].distance, 11.0)
        self.assertEqual(actual[1].distance_in_steps, 11.0)
        self.assertEqual(actual[1].distance_in_world, 1.5)
        self.assertEqual(actual[1].held, False)
        self.assertEqual(actual[1].mass, 12.34)
        self.assertEqual(actual[1].material_list, ["METAL", "PLASTIC"])
        self.assertEqual(actual[1].position, {"x": 1, "y": 2, "z": 3})
        self.assertEqual(actual[1].rotation, {"x": 1, "y": 2, "z": 3})
        self.assertEqual(actual[1].shape, 'shape2')
        self.assertEqual(actual[1].state_list, [])
        self.assertEqual(actual[1].texture_color_list, ['c2', 'c3'])
        self.assertEqual(actual[1].visible, True)
        self.assertEqual(actual[1].associated_with_agent, "agent_test")
        self.assertEqual(
            actual[1].simulation_agent_held_object, "")
        self.assertEqual(
            actual[1].simulation_agent_is_holding_held_object,
            False)

    def test_retrieve_object_list_with_states(self):
        scene_config = {
            'name': 'test name',
            'objects': [{
                'id': 'testId1',
                'type': "ball",
                'states': [['a', 'b'], ['c', 'd']]
            }]
        }

        scene_config = SceneConfiguration(**scene_config)

        mock_scene_event_data = self.create_retrieve_object_list_scene_event()

        mock_event = self.create_mock_scene_event(mock_scene_event_data)

        scene_event = SceneEvent(self._config, scene_config, mock_event, 0)
        actual = scene_event.object_list

        mock_scene_event_data = self.create_retrieve_object_list_scene_event()
        mock_event = self.create_mock_scene_event(mock_scene_event_data)

        scene_event = SceneEvent(self._config, scene_config, mock_event, 0)

        actual = scene_event.object_list
        self.assertEqual(len(actual), 2)

        self.assertEqual(actual[0].uuid, "testId1")
        self.assertEqual(actual[0].state_list, ['a', 'b'])

        self.assertEqual(actual[1].uuid, "testId2")
        self.assertEqual(actual[1].state_list, [])

    def test_retrieve_object_list_with_config_metadata_oracle(self):
        self._config.set_metadata_tier('oracle')
        mock_scene_event_data = self.create_retrieve_object_list_scene_event()
        mock_event = self.create_mock_scene_event(mock_scene_event_data)
        scene_event = SceneEvent(
            self._config,
            SceneConfiguration(),
            mock_event,
            0)
        actual = scene_event.object_list
        self.assertEqual(len(actual), 3)

        self.assertEqual(actual[0].uuid, "testId1")
        self.assertEqual(actual[0].segment_color, {
            "r": 12,
            "g": 34,
            "b": 56
        })
        self.assertEqual(actual[0].dimensions, [])
        self.assertEqual(actual[0].direction, {
            "x": 0,
            "y": 0,
            "z": 0
        })
        self.assertEqual(actual[0].distance, 0)
        self.assertEqual(actual[0].distance_in_steps, 0)
        self.assertEqual(actual[0].distance_in_world, 0)
        self.assertEqual(actual[0].held, True)
        self.assertEqual(actual[0].mass, 1)
        self.assertEqual(actual[0].material_list, [])
        self.assertEqual(actual[0].position, {"x": 1, "y": 1, "z": 2})
        self.assertEqual(actual[0].rotation, {"x": 1, "y": 2, "z": 3})
        self.assertEqual(actual[0].shape, 'shape1')
        self.assertEqual(actual[0].state_list, [])
        self.assertEqual(actual[0].texture_color_list, ['c1'])
        self.assertEqual(actual[0].visible, True)
        self.assertEqual(actual[0].associated_with_agent, "")
        self.assertEqual(
            actual[0].simulation_agent_held_object, "")
        self.assertEqual(
            actual[0].simulation_agent_is_holding_held_object,
            False)

        self.assertEqual(actual[1].uuid, "testId2")
        self.assertEqual(actual[1].segment_color, {
            "r": 98,
            "g": 76,
            "b": 54
        })
        self.assertEqual(
            actual[1].dimensions, [
                "p1", "p2", "p3", "p4", "p5", "p6", "p7", "p8"])
        self.assertEqual(actual[1].direction, {
            "x": 90,
            "y": -30,
            "z": 0
        })
        self.assertEqual(actual[1].distance, 11.0)
        self.assertEqual(actual[1].distance_in_steps, 11.0)
        self.assertEqual(actual[1].distance_in_world, 1.5)
        self.assertEqual(actual[1].held, False)
        self.assertEqual(actual[1].mass, 12.34)
        self.assertEqual(actual[1].material_list, ["METAL", "PLASTIC"])
        self.assertEqual(actual[1].position, {"x": 1, "y": 2, "z": 3})
        self.assertEqual(actual[1].rotation, {"x": 1, "y": 2, "z": 3})
        self.assertEqual(actual[1].shape, 'shape2')
        self.assertEqual(actual[1].state_list, [])
        self.assertEqual(actual[1].texture_color_list, ['c2', 'c3'])
        self.assertEqual(actual[1].visible, True)
        self.assertEqual(actual[1].associated_with_agent, "agent_test")
        self.assertEqual(
            actual[1].simulation_agent_held_object, "")
        self.assertEqual(
            actual[1].simulation_agent_is_holding_held_object,
            False)

        self.assertEqual(actual[2].uuid, "testId3")
        self.assertEqual(actual[2].segment_color, {
            "r": 101,
            "g": 102,
            "b": 103
        })
        self.assertEqual(
            actual[2].dimensions, [
                "pA", "pB", "pC", "pD", "pE", "pF", "pG", "pH"])
        self.assertEqual(actual[2].direction, {
            "x": -90,
            "y": 180,
            "z": 270
        })
        self.assertEqual(actual[2].distance, 20.0)
        self.assertEqual(actual[2].distance_in_steps, 20.0)
        self.assertEqual(actual[2].distance_in_world, 2.5)
        self.assertEqual(actual[2].held, False)
        self.assertEqual(actual[2].mass, 34.56)
        self.assertEqual(actual[2].material_list, ["WOOD"])
        self.assertEqual(actual[2].position, {"x": -3, "y": -2, "z": -1})
        self.assertEqual(actual[2].rotation, {"x": 11, "y": 12, "z": 13})
        self.assertEqual(actual[2].shape, 'shape3')
        self.assertEqual(actual[2].state_list, [])
        self.assertEqual(actual[2].texture_color_list, [])
        self.assertEqual(actual[2].visible, False)
        self.assertEqual(actual[2].associated_with_agent, "")
        self.assertEqual(
            actual[2].simulation_agent_held_object, "held_test")
        self.assertEqual(
            actual[2].simulation_agent_is_holding_held_object,
            True)

    def test_retrieve_object_list_with_config_metadata_level2(self):
        self._config.set_metadata_tier('level2')
        mock_scene_event_data = self.create_retrieve_object_list_scene_event()
        mock_event = self.create_mock_scene_event(mock_scene_event_data)
        scene_event = SceneEvent(
            self._config,
            SceneConfiguration(),
            mock_event,
            0)
        actual = scene_event.object_list
        self.assertEqual(len(actual), 3)

    def test_retrieve_object_list_with_config_metadata_level1(self):
        self._config.set_metadata_tier('level1')
        mock_scene_event_data = self.create_retrieve_object_list_scene_event()
        mock_event = self.create_mock_scene_event(mock_scene_event_data)
        scene_event = SceneEvent(
            self._config,
            SceneConfiguration(),
            mock_event,
            0)
        actual = scene_event.object_list
        self.assertEqual(len(actual), 3)

    def test_retrieve_object_list_with_config_metadata_none(self):
        self._config.set_metadata_tier('none')
        mock_scene_event_data = self.create_retrieve_object_list_scene_event()
        mock_event = self.create_mock_scene_event(mock_scene_event_data)
        scene_event = SceneEvent(
            self._config,
            SceneConfiguration(),
            mock_event,
            0)
        actual = scene_event.object_list
        self.assertEqual(len(actual), 3)

    def test_get_restrictions_none(self):
        self._config.set_metadata_tier('none')
        coh = ControllerOutputHandler(self._config)
        restricted = False
        restrictions = coh.get_restrictions(restricted, MetadataTier.NONE)
        self.assertFalse(any(restrictions))

        restricted = True
        restrictions = coh.get_restrictions(restricted, MetadataTier.NONE)
        self.assertTrue(all(restrictions))

    def test_get_restrictions_level1(self):
        self._config.set_metadata_tier('level1')
        coh = ControllerOutputHandler(self._config)
        restricted = False
        restrictions = coh.get_restrictions(restricted, MetadataTier.LEVEL_1)
        self.assertFalse(any(restrictions))

        restricted = True
        restrictions = coh.get_restrictions(restricted, MetadataTier.LEVEL_1)
        self.assertTrue(any(restrictions))
        self.assertFalse(restrictions[0])
        self.assertTrue(restrictions[1])
        self.assertTrue(restrictions[2])

    def test_get_restrictions_level2(self):
        self._config.set_metadata_tier('level2')
        coh = ControllerOutputHandler(self._config)
        restricted = False
        restrictions = coh.get_restrictions(restricted, MetadataTier.LEVEL_2)
        self.assertFalse(any(restrictions))

        restricted = True
        restrictions = coh.get_restrictions(restricted, MetadataTier.LEVEL_2)
        self.assertTrue(any(restrictions))
        self.assertFalse(restrictions[0])
        self.assertFalse(restrictions[1])
        self.assertTrue(restrictions[2])

    def test_get_restrictions_oracle(self):
        self._config.set_metadata_tier('oracle')
        coh = ControllerOutputHandler(self._config)
        restricted = False
        restrictions = coh.get_restrictions(restricted, MetadataTier.ORACLE)
        self.assertFalse(any(restrictions))

        restricted = True
        restrictions = coh.get_restrictions(restricted, MetadataTier.ORACLE)
        self.assertFalse(any(restrictions))
        self.assertFalse(restrictions[0])
        self.assertFalse(restrictions[1])
        self.assertFalse(restrictions[2])


if __name__ == '__main__':
    unittest.main()



# This is additional documentation for logging configuration dictionaries.
# Documentation on the format can be found here:
# https://docs.python.org/3/library/logging.config.html#logging-config-dictschema
#
#
# Basic structure of logging.
#   Code should get and use a logger determined by the class name:
#     logger = logging.getLogger(__name__)
#   Loggers are configured here to pass log data zero or more handlers
#     Loggers have a level which filters out any logs below that level
#   Handlers determine where to send the data (console, files, sockets)
#     Handlers have a level which filters out any logs below that level
#     Handlers can reference a formatter
#   Formatters determine the format of the log message
#   Note: For a message to be logged, its level must be equal to or above
#     both the logger's level and the handler's level
#
# "root" is Essentially the default.  Anything that isn't defined elsewhere,
# will take these settings
#
# "loggers" Sets all machine_common_sense loggers to these settings and
# logs them into the listed handlers
#
# "handlers" take log messages and put them somewhere (console, files,
# sockets, syslog, smtp, http)
# See https://docs.python.org/3/library/logging.handlers.html for
# all the different built in handlers
#
# The file handlers have a nice feature to rotate log files so you
# don't use up all your disk space (maxBytes and backupCount).
#
# creating separate file handlers can be useful, but we don't seem to have
# enough differentiation between debug and info yet to make this useful
#
# Formatters just specify the format of the log message.  Sometimes its
# useful to have a lot of detail.  Sometimes it more annoying than useful.
# Simply change which formatter you want in the handlers
# section.

import ast
import copy
import logging
import os
from typing import List, Union

logger = logging.getLogger(__name__)

TRACE = 5


class LoggingConfig():

    @staticmethod
    def init_logging(log_config=None,
                     log_config_file="log.config.user.py"):
        """
        Initializes logging system.  Attempts to read file first,
        which should not be checked in and each user can have their own.
        If user file doesn't exist, then there is a base config that
        should be read.
        """
        LoggingConfig._add_trace()
        init_message = ""
        if (os.path.exists(log_config_file)):
            with open(log_config_file, "r") as data:
                log_config = ast.literal_eval(data.read())
                init_message = f"Loaded logging config from {log_config_file}"
        elif log_config is not None:
            init_message = "Loaded provided logging config dictionary"
        if (log_config is None):
            log_config = LoggingConfig.get_default_console_config()
            init_message = "Loaded default logging config"

        if (not os.path.exists("logs")):
            os.mkdir("logs")
        logging.config.dictConfig(log_config)
        logger.info(init_message)

    @staticmethod
    def _add_trace():
        logging.addLevelName(TRACE, "TRACE")

        def trace(self, message, *args, **kws):
            if self.isEnabledFor(TRACE):
                self._log(TRACE, message, args, **kws)
        logging.Logger.trace = trace
        logging.TRACE = TRACE

    @staticmethod
    def get_default_console_config():
        return LoggingConfig.get_configurable_logging_config(
            log_level='DEBUG',
            logger_names=['machine_common_sense'],
            console=True, debug_file=False, info_file=False,
            log_file_name="mcs",
            console_format='brief'
        )

    @staticmethod
    def get_errors_only_console_config():
        return LoggingConfig.get_configurable_logging_config(
            log_level='ERROR',
            logger_names=['machine_common_sense'],
            console=True, debug_file=False, info_file=False,
            log_file_name="mcs", file_format='precise',
            console_format='brief')

    @staticmethod
    def get_no_logging_config():
        return {
            "version": 1,
            "loggers": {},
            "handlers": {}
        }

    def get_dev_logging_config():
        '''Note: This logging configuration needs the log directory to be
        created relative to the current working directory of the python
        execution.
        '''
        return LoggingConfig.get_configurable_logging_config(
            log_level='DEBUG',
            logger_names=['machine_common_sense'],
            console=True, debug_file=True, info_file=False,
            log_file_name="mcs", file_format='precise',
            console_format='brief')

    @staticmethod
    def get_configurable_logging_config(
            log_level: str = 'DEBUG',
            logger_names: Union[List[str], str] = None,
            console: bool = True,
            debug_file: bool = True,
            info_file: bool = False,
            log_file_name: str = "mcs",
            file_format: str = 'precise',
            console_format: str = 'brief',
            root_log_level: str = 'WARN'):
        """[summary]

        Args:
            log_level (str, optional): The log level for all loggers including
                root.  Can be 'ERROR', 'WARNING', 'INFO', 'DEBUG', or 'TRACE'.
                Defaults to 'DEBUG'.
            logger_names (Union[List[str], str], optional): Names of the
                loggers to create.  Each logger specifies the level for all
                packages or modules under it.
                Defaults to ['machine_common_sense'].
            console (bool, optional): Whether the console logger is on or off.
                Defaults to True.
            debug_file (bool, optional): Whether logs should be written to the
                debug log file. This requires the 'log' directory to exist.
                Defaults to True.
            info_file (bool, optional): Whether logs should be written to the
                info log file. This requires the 'log' directory to exist.
                Defaults to False.
            log_file_name (str, optional): The base name of the log file.  This
                is usually an abbreviate for the project or product.
                Defaults to "mcs".
            file_format (str, optional): Format of the log output for log
                files. Options: 'brief', 'precise', 'full'
                Defaults to 'precise'.
            console_format (str, optional): Format of the log output for
                console logging. Options: 'brief', 'precise', 'full'.
                Defaults to 'brief'.

        Returns:
            A dictionary containing a logging configuration created from the
            parameters.
        """

        if logger_names is None:
            logger_names = ['machine_common_sense']
        logger_names = logger_names if isinstance(
            logger_names, list) else [logger_names]
        handler_tags = []

        handlers = {}
        if console:
            handler_tags.append("console")
            handlers['console'] = {
                "class": "logging.StreamHandler",
                "formatter": console_format,
                "level": log_level,
                "stream": "ext://sys.stdout"
            }
        if debug_file:
            handler_tags.append("debug-file")
            handlers["debug-file"] = {
                "level": "DEBUG",
                "class": "logging.handlers.RotatingFileHandler",
                "formatter": file_format,
                "filename": f"logs/{log_file_name}.debug.log",
                "maxBytes": 10240000,
                "backupCount": 3
            }
        if info_file:
            handler_tags.append("info-file")
            handlers["info-file"] = {
                "level": "INFO",
                "class": "logging.handlers.RotatingFileHandler",
                "formatter": file_format,
                "filename": f"logs/{log_file_name}.debug.log",
                "maxBytes": 10240000,
                "backupCount": 3
            }

        logger_template = {
            "level": log_level,
            "handlers": handler_tags,
            "propagate": False
        }

        loggers = {}
        for name in logger_names:
            logger = copy.deepcopy(logger_template)
            loggers[name] = logger

        return {
            "version": 1,
            "root": {
                "level": root_log_level,
                "handlers": handler_tags,
                "propagate": False
            },
            "loggers": loggers,
            "handlers": handlers,
            "formatters": {
                "brief": {
                    "format": "%(message)s"
                },
                "precise": {
                    "format": "%(asctime)s <%(levelname)s>: %(message)s"
                },
                "full": {
                    "format": "[%(name)s] %(asctime)s <%(levelname)s>: "
                    "%(message)s"
                }
            }
        }


# Call init logging early with no config to ensure it is called.  Logging
# config can always be overwritten any time MCS is used.
LoggingConfig.init_logging(LoggingConfig.get_errors_only_console_config())



import argparse
import os

from machine_common_sense.unity_executable_provider import \
    UnityExecutableProvider

'''
Script will start Unity build with a flag to clear the addressables cache,
cache all the data in catalog and then close the application.

Usage: python cache_addressables.py <unity-build-path>
  unity-build-path: path to Unity AI2THOR build

See code details on how this works inside Unity in  the MCS AI2Thor code at
AddressablesUtil.Awake()

Normally, Unity caches what you use as you use it.  All MCS AI2Thor
executables use the same catalogs the cache is commonly shared between
executables.

'''


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--build_path',
                        help="Unity build path.",
                        required=False,
                        default=None)
    return parser.parse_args()


def run_caching_build(build_path):
    os.execl(build_path, build_path, "CACHEADDRESSABLES")


def main():
    args = parse_args()
    path = args.build_path
    if not path:
        path = UnityExecutableProvider().get_executable(force_download=True)
    run_caching_build(path)


if __name__ == '__main__':
    main()



import logging
import pathlib
from abc import abstractmethod

import numpy as np
import PIL

from .controller_events import (AbstractControllerSubscriber, AfterStepPayload,
                                BasePostActionEventPayload,
                                ControllerEventPayload, EndScenePayload,
                                StartScenePayload)
from .plotter import TopDownPlotter
from .recorder import VideoRecorder

logger = logging.getLogger(__name__)


DEPTH_HSV_BANDS = 400.0


def convert_depth_to_hsv(
    depth_float_array: np.array,
    clipping_plane_far: float
) -> np.array:
    # Split the depth data into multiple color bands. Each band corresponds to
    # the full hue spectrum, from red to purple, at a specific saturation and
    # value. Near bands have lower saturation (closer to white); far bands have
    # lower value (closer to black). Don't use saturation or value lower than
    # 30, or else the color will look too close to white or black.
    band_size = clipping_plane_far / DEPTH_HSV_BANDS
    mid_sat = band_size * int(DEPTH_HSV_BANDS / 2.0)
    mid_val = clipping_plane_far - (band_size * int(DEPTH_HSV_BANDS / 2.0))
    shape = depth_float_array.shape
    depth_hsv_array = np.zeros((shape[0], shape[1], 3), dtype=np.uint8)
    hue = (np.mod(depth_float_array, band_size) / band_size * 256)
    sat = np.where(
        depth_float_array < mid_sat,
        55 + np.floor(depth_float_array / band_size),
        255
    )
    val = np.where(
        depth_float_array >= mid_val,
        255 - np.floor((depth_float_array - mid_val) / band_size),
        255
    )
    # Note that PIL images use HSV ints between 0 and 255 inclusive.
    depth_hsv_array[:shape[0], :shape[1], 0] = hue
    depth_hsv_array[:shape[0], :shape[1], 1] = sat
    depth_hsv_array[:shape[0], :shape[1], 2] = val
    return depth_hsv_array


def convert_depth_to_image(
    depth_float_array: np.array,
    clipping_plane_far: float
) -> PIL.Image:
    '''Convert the given depth float array into a depth image, then return the
    image.'''
    # Convert the depth array from a 2D array of floats, ranging from 0 to the
    # camera's far clipping plane, to a 3D array of HSV tuples.
    depth_hsv_array = convert_depth_to_hsv(
        depth_float_array,
        clipping_plane_far
    )
    image = PIL.Image.fromarray(depth_hsv_array, mode='HSV')
    image = image.convert('RGB')
    return image


class AbstractImageEventHandler(AbstractControllerSubscriber):
    '''Abstract class for handling saving different images based on controller
    events.  This class assumes images should be saved on the start of the
    scene and on each step.  The implementation should create the image (PIL)
    and then call _do_save_image to save the file.
    '''

    def on_start_scene(self, payload: ControllerEventPayload):
        self._save_image(payload)

    def on_after_step(self, payload: ControllerEventPayload):
        self._save_image(payload)

    @abstractmethod
    def _save_image(self, payload):
        pass

    def _get_suffix(self, payload, index: int):
        if payload.step_number == 0:
            step_plus_substep_index = 0
        else:
            step_plus_substep_index = (
                ((payload.step_number - 1) *
                    len(payload.step_metadata.events)) +
                (index + 1)
            )
        return f"_{str(step_plus_substep_index)}.png"

    def _do_save_image(self, payload, index: int, data: PIL.Image, name: str):
        if (payload.output_folder):
            suffix = self._get_suffix(payload, index)
            data.save(fp=payload.output_folder +
                      name + suffix)


class AbstractVideoEventHandler(AbstractControllerSubscriber):

    '''Abstract class for handling saving different videos based on controller
        events.  This method provides some commonly used support methods, but
        leaves most of the work up to the implementation.  Typically, the
        implmentation should call create_video_recorder.  Images can be added
        to the video based on the implementations needs and at the end the
        recorder needs to be finished.
    '''

    PLACEHOLDER = 'placeholder'
    VISUAL = 'visual'
    DEPTH = 'depth'
    SEGMENTATION = 'segmentation'
    TOPDOWN = 'topdown'

    def get_basename(self, payload: ControllerEventPayload):
        timestamp = payload.timestamp
        scene_name = payload.scene_config.name.replace('json', '')
        return '_'.join(
            [payload.config.get_evaluation_name(),
             payload.config.get_metadata_tier().value,
             payload.config.get_team(),
             scene_name,
             AbstractVideoEventHandler.PLACEHOLDER, timestamp]) + '.mp4'

    def get_scene_name(self, payload: ControllerEventPayload):
        scene_name = payload.scene_config.name.replace('json', '')

        # strip prefix in scene_name
        if '/' in scene_name:
            scene_name = scene_name.rsplit('/', 1)[1]
        return scene_name

    def create_video_recorder(self, payload, file_tag):
        basename = self.get_basename(payload)
        video_filename = basename.replace(
            AbstractVideoEventHandler.PLACEHOLDER,
            file_tag
        )
        output_folder = pathlib.Path(payload.output_folder)
        vid_path = output_folder / video_filename
        fps = payload.step_output.physics_frames_per_second
        return VideoRecorder(vid_path, fps=fps)

    def _get_filename_without_timestamp(self, filepath: pathlib.Path):
        return filepath.stem[:-16] + filepath.suffix


class SceneImageEventHandler(AbstractImageEventHandler):
    '''
    Writes images for the visual output or scene output
    '''

    def _save_image(self, payload):
        for index, scene_image in enumerate(payload.step_output.image_list):
            self._do_save_image(payload, index, scene_image, 'frame_image')


class DepthImageEventHandler(AbstractImageEventHandler):
    '''
    writes images for depth map
    '''

    def _save_image(self, payload):
        for index, depth_float_array in enumerate(
                payload.step_output.depth_map_list):
            depth_image = convert_depth_to_image(
                depth_float_array,
                payload.step_output.camera_clipping_planes[1]
            )
            self._do_save_image(payload, index, depth_image, 'depth_map')


class ObjectMaskImageEventHandler(AbstractImageEventHandler):
    '''
    writes images for object mask
    '''

    def _save_image(self, payload):
        for index, object_mask in enumerate(
                payload.step_output.object_mask_list):
            self._do_save_image(payload, index, object_mask, 'object_mask')


class ImageVideoEventHandler(AbstractVideoEventHandler):
    '''
    writes videos for visual images
    '''

    def on_start_scene(self, payload: ControllerEventPayload):
        self.__recorder = self.create_video_recorder(
            payload, AbstractVideoEventHandler.VISUAL)
        self.save_video_for_step(payload)

    def on_after_step(self, payload: ControllerEventPayload):
        self.save_video_for_step(payload)

    def save_video_for_step(self, payload: ControllerEventPayload):
        for scene_image in payload.step_output.image_list:
            self.__recorder.add(scene_image)

    def on_end_scene(self, payload: ControllerEventPayload):
        self.__recorder.finish()


class UnityTopdownCameraCombinerEventHandler(AbstractVideoEventHandler):
    keyword = "topdown"

    def on_start_scene(self, payload: StartScenePayload):
        self.__recorder = self.create_video_recorder(
            payload, AbstractVideoEventHandler.TOPDOWN)
        self.__plotter = TopDownPlotter(
            team=payload.config.get_team(),
            scene_config=payload.scene_config
        )
        self.folder = pathlib.Path(payload.output_folder)
        self.write_image(payload)

    def write_image(self, payload: BasePostActionEventPayload):
        image_file = (
            self.folder / pathlib.Path(
                f"{payload.scene_config.name}_{self.keyword}_"
                f"{payload.step_number}.png"))
        image = PIL.Image.open(image_file.as_posix())
        self.__recorder.add(image)

    def on_after_step(self, payload: AfterStepPayload):
        self.write_image(payload)

    def on_end_scene(self, payload: EndScenePayload):
        self.__recorder.finish()


class TopdownVideoEventHandler(AbstractVideoEventHandler):
    '''
    writes top down video
    '''

    def on_start_scene(self, payload: BasePostActionEventPayload):
        self.__recorder = self.create_video_recorder(
            payload, AbstractVideoEventHandler.TOPDOWN)
        self.__plotter = TopDownPlotter(
            team=payload.config.get_team(),
            scene_config=payload.scene_config
        )
        self.save_video_for_step(payload)

    def on_after_step(self, payload: BasePostActionEventPayload):
        self.save_video_for_step(payload)

    def save_video_for_step(self, payload: BasePostActionEventPayload):
        for _ in payload.step_metadata.events:
            # The plotter used to be inside the for loop the same as
            # image_recorder, but it seems like it would only plot one per
            # step.

            target_ids = []
            metadata = (payload.goal.metadata or {}) if payload.goal else {}
            # Different goal categories may use different property names
            target_names = ['target', 'targets', 'target_1', 'target_2']
            for target_name in target_names:
                # Some properties may be dicts, and some may be lists of dicts
                targets = metadata.get(target_name) or []
                targets = targets if isinstance(targets, list) else [targets]
                for target in targets:
                    if target.get('id'):
                        target_ids.append(target.get('id'))

            plot = self.__plotter.plot(payload.step_metadata,
                                       payload.step_number,
                                       target_ids)
            self.__recorder.add(plot)

    def on_end_scene(self, payload: BasePostActionEventPayload):
        self.__recorder.finish()


class DepthVideoEventHandler(AbstractVideoEventHandler):
    '''
    writes video of depth maps
    '''

    def on_start_scene(self, payload: BasePostActionEventPayload):
        self.__recorder = self.create_video_recorder(
            payload, AbstractVideoEventHandler.DEPTH)
        self.save_video_for_step(payload)

    def on_after_step(self, payload: BasePostActionEventPayload):
        self.save_video_for_step(payload)

    def save_video_for_step(self, payload: BasePostActionEventPayload):
        for depth_float_array in payload.step_output.depth_map_list:
            depth_image = convert_depth_to_image(
                depth_float_array,
                payload.step_output.camera_clipping_planes[1]
            )
            self.__recorder.add(depth_image)

    def on_end_scene(self, payload: BasePostActionEventPayload):
        self.__recorder.finish()


class SegmentationVideoEventHandler(AbstractVideoEventHandler):
    '''
    writes video for segmentation or object mask images
    '''

    def on_start_scene(self, payload: BasePostActionEventPayload):
        self.__recorder = self.create_video_recorder(
            payload, AbstractVideoEventHandler.SEGMENTATION)
        self.save_video_for_step(payload)

    def on_after_step(self, payload: BasePostActionEventPayload):
        self.save_video_for_step(payload)

    def save_video_for_step(self, payload: BasePostActionEventPayload):
        for object_mask in payload.step_output.object_mask_list:
            self.__recorder.add(object_mask)

    def on_end_scene(self, payload: BasePostActionEventPayload):
        self.__recorder.finish()



import datetime
import enum
from abc import ABC
from typing import Dict, Optional, Union

from ai2thor.server import Event, MultiAgentEvent

from .config_manager import BaseModel, ConfigManager, SceneConfiguration
from .goal_metadata import GoalMetadata
from .step_metadata import StepMetadata


class EventType(enum.Enum):
    '''
    Enum for the possible events the controller can send to subscribers
    '''
    ON_START_SCENE = enum.auto()
    ON_BEFORE_STEP = enum.auto()
    ON_AFTER_STEP = enum.auto()
    ON_END_SCENE = enum.auto()


class BaseEventPayload(BaseModel):
    step_number: int
    config: ConfigManager
    scene_config: Optional[SceneConfiguration]


class BasePostActionEventPayload(BaseEventPayload):
    output_folder: Optional[str]
    timestamp: str
    wrapped_step: dict
    step_metadata: Union[Event, MultiAgentEvent]
    step_output: StepMetadata
    restricted_step_output: StepMetadata
    goal: GoalMetadata


class StartScenePayload(BasePostActionEventPayload):
    ...


class AfterStepPayload(BasePostActionEventPayload):
    ai2thor_action: str
    step_params: dict
    action_kwargs: dict


class BeforeStepPayload(BaseEventPayload):
    action: str
    goal: GoalMetadata
    habituation_trial: Optional[int]


class EndScenePayload(BaseEventPayload):
    rating: Optional[str]
    score: Optional[float]
    report: Optional[dict]


class ControllerEventPayload(BaseEventPayload):
    '''
    Payload class for controller events.  This class will be redone soon.
    '''

    def __init__(self, output_folder: str, config: ConfigManager,
                 step_number: int, scene_config: SceneConfiguration,
                 habituation_trial: int, goal: Dict):
        self.timestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        self.output_folder = output_folder
        self.config = config
        self.step_number = step_number
        self.scene_config = scene_config
        self.habituation_trial = habituation_trial
        self.goal = goal

    def get_output_folder(self) -> str:
        return self.output_folder

    def get_config(self) -> ConfigManager:
        return self.config

    def get_timestamp(self) -> str:
        return self.timestamp

    def get_step_number(self) -> int:
        return self.step_number

    def get_scene_config(self) -> SceneConfiguration:
        self.scene_config

    def get_habituation_trial(self) -> int:
        return self.habituation_trial

    def get_goal(self) -> GoalMetadata:
        return self.goal


class AbstractControllerSubscriber(ABC):
    '''
    Abstract class for controller event subscribers.  Subscriber
    implementations should override at least one of the event classes.
    '''

    def __init__(self):
        self._switcher = {
            EventType.ON_START_SCENE: self.on_start_scene,
            EventType.ON_BEFORE_STEP: self.on_before_step,
            EventType.ON_AFTER_STEP: self.on_after_step,
            EventType.ON_END_SCENE: self.on_end_scene
        }

    def on_event(self, type: EventType,
                 payload: ControllerEventPayload):
        self._switcher.get(
            type,
            "default")(
            payload)

    def on_start_scene(self, payload: StartScenePayload):
        pass

    def on_before_step(self, payload: BeforeStepPayload):
        pass

    def on_after_step(self, payload: AfterStepPayload):
        pass

    def on_end_scene(self, payload: EndScenePayload):
        pass



import copy
import textwrap
import unittest

import machine_common_sense as mcs


class TestStepMetadata(unittest.TestCase):

    str_output = '''    {
        "action_list": [],
        "camera_aspect_ratio": [0.0,0.0],
        "camera_clipping_planes": [0.0,0.0],
        "camera_field_of_view": 0.0,
        "camera_height": 0.0,
        "depth_map_list": [],
        "goal": {
            "action_list": null,
            "category": "",
            "description": "",
            "habituation_total": 0,
            "last_preview_phase_step": 0,
            "last_step": null,
            "metadata": {},
            "steps_allowed_in_lava": 0,
            "triggered_by_target_sequence": null
        },
        "habituation_trial": null,
        "haptic_feedback": {},
        "head_tilt": 0.0,
        "holes": [],
        "image_list": [],
        "lava": [],
        "object_list": [],
        "object_mask_list": [],
        "performer_radius": 0.0,
        "performer_reach": 0.0,
        "physics_frames_per_second": 0,
        "position": {},
        "resolved_object": "",
        "resolved_receptacle": "",
        "return_status": "UNDEFINED",
        "reward": 0,
        "room_dimensions": {},
        "rotation": 0.0,
        "segmentation_colors": [],
        "step_number": 0,
        "steps_on_lava": 0,
        "structural_object_list": [],
        "triggered_by_sequence_incorrect": false
    }'''

    str_output_segment_map_ints = '''    {
        "action_list": [],
        "camera_aspect_ratio": [0.0,0.0],
        "camera_clipping_planes": [0.0,0.0],
        "camera_field_of_view": 0.0,
        "camera_height": 0.0,
        "depth_map_list": [],
        "goal": {
            "action_list": null,
            "category": "",
            "description": "",
            "habituation_total": 0,
            "last_preview_phase_step": 0,
            "last_step": null,
            "metadata": {},
            "steps_allowed_in_lava": 0,
            "triggered_by_target_sequence": null
        },
        "habituation_trial": null,
        "haptic_feedback": {},
        "head_tilt": 0.0,
        "holes": [],
        "image_list": [],
        "lava": [],
        "object_list": [],
        "object_mask_list": [],
        "performer_radius": 0.0,
        "performer_reach": 0.0,
        "physics_frames_per_second": 0,
        "position": {},
        "resolved_object": "",
        "resolved_receptacle": "",
        "return_status": "UNDEFINED",
        "reward": 0,
        "room_dimensions": {},
        "rotation": 0.0,
        "segmentation_colors": [],
        "step_number": 0,
        "steps_on_lava": 0,
        "structural_object_list": [],
        "triggered_by_sequence_incorrect": false,
        "segment_map": {
            "0": {
                "r": 218,
                "g": 65,
                "b": 21
            }
        }
    }'''

    @classmethod
    def setUpClass(cls):
        cls.step_metadata = mcs.StepMetadata()

    @classmethod
    def tearDownClass(cls):
        # nothing to do here
        pass

    def test_action_list(self):
        self.assertFalse(self.step_metadata.action_list)
        self.assertIsInstance(self.step_metadata.action_list, list)

    def test_camera_aspect_ratio(self):
        self.assertIsInstance(self.step_metadata.camera_aspect_ratio, tuple)

    def test_camera_clipping_planes(self):
        self.assertIsInstance(
            self.step_metadata.camera_clipping_planes, tuple)

    def test_camera_field_of_view(self):
        self.assertIsInstance(self.step_metadata.camera_field_of_view, float)

    def test_camera_height(self):
        self.assertIsInstance(self.step_metadata.camera_height, float)

    def test_depth_map_list(self):
        self.assertFalse(self.step_metadata.depth_map_list)
        self.assertIsInstance(self.step_metadata.depth_map_list, list)

    def test_goal(self):
        self.assertIsInstance(self.step_metadata.goal, mcs.GoalMetadata)

    def test_habituation_trial(self):
        self.assertEqual(self.step_metadata.habituation_trial, None)

    def test_head_tilt(self):
        self.assertAlmostEqual(self.step_metadata.head_tilt, 0.0)
        self.assertIsInstance(self.step_metadata.head_tilt, float)

    def test_holes(self):
        self.assertEqual(self.step_metadata.holes, [])
        self.assertIsInstance(self.step_metadata.holes, list)

    def test_image_list(self):
        self.assertFalse(self.step_metadata.image_list)
        self.assertIsInstance(self.step_metadata.image_list, list)

    def test_lava(self):
        self.assertEqual(self.step_metadata.lava, [])
        self.assertIsInstance(self.step_metadata.lava, list)

    def test_object_list(self):
        self.assertFalse(self.step_metadata.object_list)
        self.assertIsInstance(self.step_metadata.object_list, list)

    def test_object_mask_list(self):
        self.assertFalse(self.step_metadata.object_mask_list)
        self.assertIsInstance(self.step_metadata.object_mask_list, list)

    def test_performer_radius(self):
        self.assertEqual(self.step_metadata.performer_radius, 0.0)
        self.assertIsInstance(self.step_metadata.performer_radius, float)

    def test_performer_reach(self):
        self.assertEqual(self.step_metadata.performer_reach, 0.0)
        self.assertIsInstance(self.step_metadata.performer_reach, float)

    def test_position(self):
        self.assertIsInstance(self.step_metadata.position, dict)

    def test_resolved_object(self):
        self.assertEqual(self.step_metadata.resolved_object, '')
        self.assertIsInstance(self.step_metadata.resolved_object, str)

        self.step_metadata.resolved_object = 'testResolvedId'
        self.assertEqual(self.step_metadata.resolved_object, 'testResolvedId')
        self.assertIsInstance(self.step_metadata.resolved_object, str)

        self.step_metadata.resolved_object = 'testResolved'
        self.assertNotEqual(
            self.step_metadata.resolved_object,
            'testResolvedId')
        self.assertIsInstance(self.step_metadata.resolved_object, str)

        self.step_metadata.resolved_object = ''

    def test_resolved_receptacle(self):
        self.assertEqual(self.step_metadata.resolved_receptacle, '')
        self.assertIsInstance(self.step_metadata.resolved_receptacle, str)

        self.step_metadata.resolved_receptacle = 'testResolvedId'
        self.assertEqual(
            self.step_metadata.resolved_receptacle,
            'testResolvedId')
        self.assertIsInstance(self.step_metadata.resolved_receptacle, str)

        self.step_metadata.resolved_receptacle = 'testResolved'
        self.assertNotEqual(
            self.step_metadata.resolved_receptacle,
            'testResolvedId')
        self.assertIsInstance(self.step_metadata.resolved_receptacle, str)

        self.step_metadata.resolved_receptacle = ''

    def test_return_status(self):
        self.assertEqual(
            self.step_metadata.return_status,
            mcs.ReturnStatus.UNDEFINED.value)
        self.assertIsInstance(self.step_metadata.return_status, str)

    def test_reward(self):
        self.assertEqual(self.step_metadata.reward, 0)
        self.assertIsInstance(self.step_metadata.reward, int)

    def test_room_dimensions(self):
        self.assertIsInstance(self.step_metadata.room_dimensions, dict)

    def test_rotation(self):
        self.assertAlmostEqual(self.step_metadata.rotation, 0.0)
        self.assertIsInstance(self.step_metadata.rotation, float)

    def test_segmentation_colors(self):
        self.assertEqual(self.step_metadata.segmentation_colors, [])
        self.assertIsInstance(self.step_metadata.segmentation_colors, list)

    def test_step_number(self):
        self.assertEqual(self.step_metadata.step_number, 0)
        self.assertIsInstance(self.step_metadata.step_number, int)

    def test_structural_object_list(self):
        self.assertFalse(self.step_metadata.structural_object_list)
        self.assertIsInstance(
            self.step_metadata.structural_object_list, list)

    def test_str(self):
        self.assertEqual(str(self.step_metadata),
                         textwrap.dedent(self.str_output))

    def test_str_int_segment_map(self):
        metadata = copy.deepcopy(self.step_metadata)
        metadata.segment_map = {
            0: {
                'r': 218,
                'g': 65,
                'b': 21
            }
        }
        self.assertEqual(str(metadata),
                         textwrap.dedent(self.str_output_segment_map_ints))

    def test_triggered_by_sequence_incorrect(self):
        self.assertEqual(
            self.step_metadata.triggered_by_sequence_incorrect, False)
        self.assertIsInstance(
            self.step_metadata.triggered_by_sequence_incorrect, bool)

    def test_copy_without_depth_or_images(self):
        data = mcs.StepMetadata(
            action_list=['action_1', 'action_2'],
            camera_aspect_ratio=[3, 2],
            camera_clipping_planes=[0, 10],
            camera_field_of_view=10,
            camera_height=2,
            depth_map_list=['depth_1', 'depth_2'],
            goal=mcs.GoalMetadata(metadata={'key': 'value'}),
            habituation_trial=1,
            head_tilt=15,
            image_list=['image_1', 'image_2'],
            object_list=[
                mcs.ObjectMetadata(uuid='object_1'),
                mcs.ObjectMetadata(uuid='object_2')
            ],
            object_mask_list=['segmentation_1', 'segmentation_2'],
            performer_radius=0.5,
            performer_reach=1,
            physics_frames_per_second=20,
            position={'x': 1, 'z': 2},
            return_status=mcs.ReturnStatus.SUCCESSFUL.value,
            resolved_object='testObjectId',
            resolved_receptacle='',
            reward=0,
            room_dimensions={'x': 10, 'y': 4, 'z': 12},
            rotation=90,
            segmentation_colors=[{'objectId': 'a', 'r': 0, 'g': 1, 'b': 2}],
            step_number=25,
            structural_object_list=[
                mcs.ObjectMetadata(uuid='structure_1'),
                mcs.ObjectMetadata(uuid='structure_2')
            ],
            triggered_by_sequence_incorrect=True
        )
        copy = data.copy_without_depth_or_images()
        # Assert are exactly equal
        self.assertEqual(data.action_list, copy.action_list)
        self.assertEqual(data.camera_aspect_ratio, copy.camera_aspect_ratio)
        self.assertEqual(
            data.camera_clipping_planes,
            copy.camera_clipping_planes
        )
        self.assertEqual(
            data.camera_field_of_view,
            copy.camera_field_of_view
        )
        self.assertEqual(data.camera_height, copy.camera_height)
        self.assertEqual(dict(data.goal), dict(copy.goal))
        self.assertEqual(data.habituation_trial, copy.habituation_trial)
        self.assertEqual(data.head_tilt, copy.head_tilt)
        self.assertEqual(
            [dict(object_data) for object_data in data.object_list],
            [dict(object_data) for object_data in copy.object_list]
        )
        self.assertEqual(data.performer_radius, copy.performer_radius)
        self.assertEqual(data.performer_reach, copy.performer_reach)
        self.assertEqual(
            data.physics_frames_per_second,
            copy.physics_frames_per_second
        )
        self.assertEqual(data.position, copy.position)
        self.assertEqual(data.return_status, copy.return_status)
        self.assertEqual(data.resolved_object, copy.resolved_object)
        self.assertEqual(data.resolved_receptacle, copy.resolved_receptacle)
        self.assertEqual(data.reward, copy.reward)
        self.assertEqual(data.room_dimensions, copy.room_dimensions)
        self.assertEqual(data.rotation, copy.rotation)
        self.assertEqual(data.segmentation_colors, copy.segmentation_colors)
        self.assertEqual(data.step_number, copy.step_number)
        self.assertEqual(
            [dict(object_data) for object_data in data.structural_object_list],
            [dict(object_data) for object_data in copy.structural_object_list]
        )
        # Assert are empty lists
        self.assertEqual(copy.depth_map_list, [])
        self.assertEqual(copy.image_list, [])
        self.assertEqual(copy.object_mask_list, [])
        # Assert are not the same instances
        self.assertNotEqual(data.goal, copy.goal)
        self.assertNotEqual(data.object_list, copy.object_list)
        self.assertNotEqual(
            data.structural_object_list,
            copy.structural_object_list
        )
        self.assertEqual(
            data.triggered_by_sequence_incorrect,
            copy.triggered_by_sequence_incorrect
        )


if __name__ == '__main__':
    unittest.main()



import argparse
import glob
import json


def main(debug_scene_folder: str, output_file_name: str):
    json_file_list = glob.glob(f"{debug_scene_folder}*_debug.json")
    json_file_list = sorted(json_file_list)
    print(f'Found {len(json_file_list)} debug files in {debug_scene_folder}')
    all_ground_truth = {}
    for json_file_name in json_file_list:
        with open(json_file_name) as json_file:
            data = json.load(json_file)
            scene_name = data['name']
            ground_truth = data['goal']['answer']['choice']
            all_ground_truth[scene_name] = ground_truth
    print(f'Saving {len(all_ground_truth.keys())} rows to {output_file_name}')
    with open(output_file_name, 'w') as output_file:
        output_file.write('scene_name,ground_truth\n')
        for scene_name, ground_truth in all_ground_truth.items():
            output_file.write(f'{scene_name},{ground_truth}\n')


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=('Extract Ground Truth'))
    parser.add_argument(
        'debug_scene_folder',
        type=str,
        help='Folder containing the MCS scene JSON debug files'
    )
    parser.add_argument(
        'output_file_name',
        type=str,
        help='Name of the output CSV file'
    )
    args = parser.parse_args()
    main(args.debug_scene_folder, args.output_file_name)



import math

from runner_script import MultipleFileRunnerScript

from machine_common_sense.controller import Controller
from machine_common_sense.controller_events import (
    AbstractControllerSubscriber, ControllerEventPayload)

MAX_DISTANCE = 0.06
MAX_DELTA_ANGLE = 11


def get_waypoint_action(delta_angle, step_metadata):
    if len(step_metadata.action_list) == 1:
        return step_metadata.action_list[0]
    if step_metadata.return_status == 'OBSTRUCTED':
        return 'OpenObject', {
            'objectImageCoordsX': 160, 'objectImageCoordsY': 120}
    if abs(delta_angle) > MAX_DELTA_ANGLE:
        return ('RotateLeft' if delta_angle > 0 else 'RotateRight', {})
    else:
        return ('MoveAhead', {})


def get_deltas(previous_output, waypoint):
    pos = previous_output.position
    rot = previous_output.rotation
    delta_x = waypoint['x'] - pos['x']
    delta_z = waypoint['z'] - pos['z']
    square_distance = delta_x * delta_x + delta_z * delta_z
    target_angle = math.degrees(math.atan2(delta_x, delta_z))
    delta_angle = (rot - target_angle + 720) % 360
    delta_angle = delta_angle if delta_angle <= 180 else delta_angle - 360
    return square_distance, delta_angle


class PathFollower(AbstractControllerSubscriber):
    waypoint_index = 0

    scene_name = ""

    def on_start_scene(self, payload: ControllerEventPayload):
        self.previous_output = payload.step_output

    def on_after_step(self, payload: ControllerEventPayload):
        self.previous_output = payload.step_output

    def init_callback(self, controller: Controller):
        controller.subscribe(self)

    def action_callback(self, scene_data, step_metadata, runner_script):
        step_metadata = self.previous_output
        if scene_data['name'] != self.scene_name:
            self.scene_name = scene_data['name']
            self.waypoint_index = 0

        path = scene_data.get('debug', {}).get('path')
        if not path:
            print("Scene did not have 'debug.path' section")
            return None, None

        if not step_metadata.position:
            print("No position provided.")
            return None, None
        waypoint = path[self.waypoint_index]
        sq_dist, delta_angle = get_deltas(step_metadata, waypoint)
        while sq_dist < MAX_DISTANCE**2:
            self.waypoint_index += 1
            if self.waypoint_index >= len(path):
                return None, None
            waypoint = path[self.waypoint_index]
            sq_dist, delta_angle = get_deltas(step_metadata, waypoint)

        return get_waypoint_action(delta_angle, step_metadata)


follower = PathFollower()


def main():
    MultipleFileRunnerScript(
        'Interactive Scenes - Follow Path',
        follower.action_callback, follower.init_callback)


if __name__ == "__main__":
    main()
